"""Models and functions used in schema documentation dialogs."""
from typing import Any, Optional

import markdown
from PySide2.QtGui import QKeySequence
from PySide2.QtWebEngineWidgets import QWebEnginePage, QWebEngineView
from PySide2.QtWidgets import QAction, QMainWindow, QWidget

from ywh2bt.core.schema.markdown import root_configuration_as_markdown
from ywh2bt.gui.widgets.typing import as_signal_instance


class SchemaDocumentationDialog(QMainWindow):
    """A window that display the schema documentation."""

    def __init__(
        self,
        parent: Optional[QWidget] = None,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """
        Initialize self.

        Args:
            parent: a parent widget
            args: extra arguments
            kwargs: extra keyword arguments
        """
        super().__init__(
            parent,
        )
        self._init_ui()

    def _init_ui(
        self,
    ) -> None:
        self._create_menus()
        self.setCentralWidget(self._create_view())
        self.statusBar().show()

    def _create_menus(
        self,
    ) -> None:
        bar = self.menuBar()

        file_menu = bar.addMenu('&File')
        file_menu.setObjectName('menu_file')
        file_menu.addAction(self._create_close_action())

    def _create_close_action(
        self,
    ) -> QAction:
        action = QAction('&Close', self)
        action.setObjectName('action_close')
        action.setShortcuts(QKeySequence.Close)
        action.setStatusTip('Close')
        as_signal_instance(action.triggered).connect(
            self._on_close_triggered,
        )
        return action

    def _on_close_triggered(
        self,
    ) -> None:
        self.close()

    def _create_view(
        self,
    ) -> QWebEngineView:
        page = QWebEnginePage(self)
        page.setHtml(
            markdown.markdown(
                root_configuration_as_markdown(),
            ),
        )
        view = QWebEngineView(self)
        view.setPage(page)
        return view



"""Models and functions used for exportable dict GUI."""
from __future__ import annotations

from functools import partial
from typing import Any, List, Optional, Set, cast

from PySide2.QtCore import QAbstractTableModel, QItemSelection, QModelIndex, Qt, Signal
from PySide2.QtWidgets import (
    QAbstractItemView,
    QHBoxLayout,
    QHeaderView,
    QLayout,
    QPushButton,
    QTableView,
    QVBoxLayout,
    QWidget,
)

from ywh2bt.core.configuration.attribute import ExportableDict
from ywh2bt.gui.widgets import constants
from ywh2bt.gui.widgets.typing import as_signal_instance


class ExportableDictWidget(QWidget):  # noqa: WPS214
    """Exportable dict GUI."""

    dataChanged: Signal = Signal(ExportableDict)  # noqa: WPS115, N815

    _table: QTableView
    _add_button: QPushButton
    _del_button: QPushButton

    def __init__(
        self,
        parent: Optional[QWidget] = None,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """
        Initialize self.

        Args:
            parent: a parent widget
            args: extra arguments
            kwargs: extra keyword arguments
        """
        super().__init__(
            parent,
            *args,
            **kwargs,
        )
        self._init_ui()

    def _init_ui(self) -> None:
        layout = QHBoxLayout(self)

        self._table = self._create_table()

        self._add_button = self._create_add_button()
        self._del_button = self._create_del_button()

        buttons_layout = self._create_buttons_layout(
            self._add_button,
            self._del_button,
        )

        layout.addWidget(self._table)
        layout.addLayout(buttons_layout)
        layout.setMargin(0)

    def _create_table(
        self,
    ) -> QTableView:
        widget = QTableView(self)
        widget.horizontalHeader().hide()
        widget.horizontalHeader().setSectionResizeMode(QHeaderView.Stretch)
        widget.setEditTriggers(
            QAbstractItemView.AllEditTriggers,
        )
        return widget

    def _create_add_button(
        self,
    ) -> QPushButton:
        widget = QPushButton(
            '+',
            self,
        )
        widget.setFixedSize(constants.SMALL_BUTTON_SIZE)
        as_signal_instance(widget.clicked).connect(
            self._on_add_button_clicked,
        )
        return widget

    def _create_del_button(
        self,
    ) -> QPushButton:
        widget = QPushButton(
            '-',
            self,
        )
        widget.setFixedSize(constants.SMALL_BUTTON_SIZE)
        widget.setEnabled(False)
        as_signal_instance(widget.clicked).connect(
            self._on_del_button_clicked,
        )
        return widget

    def _create_buttons_layout(
        self,
        *buttons: QPushButton,
    ) -> QLayout:
        layout = QVBoxLayout()
        for button in buttons:
            layout.addWidget(button, 1)
        layout.addStretch(1)
        return layout

    def set_model(
        self,
        model: ExportableDictModel,
    ) -> None:
        """
        Set the model.

        Args:
            model: an exportable dict model
        """
        signals = {
            model.dataChanged,
            model.rowsInserted,
            model.rowsRemoved,
        }
        for signal in signals:
            as_signal_instance(signal).connect(
                partial(
                    self._on_model_data_changed,
                    model,
                ),
            )
        self._table.setModel(model)
        selection_model = self._table.selectionModel()
        as_signal_instance(selection_model.selectionChanged).connect(
            self._on_selection_changed,
        )

    def _on_model_data_changed(
        self,
        model: ExportableDictModel,
        top_left: QModelIndex,
        bottom_right: QModelIndex,
        roles: List[int],
    ) -> None:
        as_signal_instance(self.dataChanged).emit(model.get_value())

    def _on_selection_changed(
        self,
        selected: QItemSelection,
        unselected: QItemSelection,
    ) -> None:
        has_selection = selected.count() != 0
        self._del_button.setEnabled(has_selection)

    def _on_add_button_clicked(
        self,
        checked: bool,
    ) -> None:
        model = self._table.model()
        model.insertRow(model.rowCount(model.index(0, 0)))
        row_count = model.rowCount(model.index(0, 0))
        self._table.selectRow(max(0, row_count - 1))

    def _on_del_button_clicked(
        self,
        checked: bool,
    ) -> None:
        selected_rows = sorted(
            self._get_select_row_indices(),
            reverse=True,
        )
        if not selected_rows:
            return
        model = self._table.model()
        for selected_row in selected_rows:
            model.removeRow(selected_row)

        row_count = model.rowCount(model.index(0, 0))
        reselect_row = min(row_count - 1, selected_rows[-1])
        self._table.selectRow(max(0, reselect_row))

    def _get_select_row_indices(
        self,
    ) -> Set[int]:
        selection_model = self._table.selectionModel()
        return set({
            index.row() for index in selection_model.selectedIndexes()
        })


class ExportableDictModel(QAbstractTableModel):
    """Exportable dict model."""

    _exportable_dict: ExportableDict[Any, Any, Any, Any]

    def __init__(
        self,
        exportable_dict: ExportableDict[Any, Any, Any, Any],
    ) -> None:
        """
        Initialize self.

        Args:
            exportable_dict: an exportable dict
        """
        super().__init__()
        self._exportable_dict = exportable_dict

    def get_value(
        self,
    ) -> Optional[ExportableDict[Any, Any, Any, Any]]:
        """
        Get the underlying exportable dict.

        Returns:
            The exportable dict
        """
        return self._exportable_dict

    def data(  # type: ignore
        self,
        index: QModelIndex,
        role: int,
    ) -> Any:
        """
        Get data at the given index for the given role.

        Args:
            index: a data index
            role: a data role

        Returns:
            The data
        """
        return_conditions = (
            self._exportable_dict is None,
            not index.isValid(),
            index.column() not in {0, 1},
        )
        if any(return_conditions):
            return None
        if role in {Qt.DisplayRole, Qt.EditRole}:
            data = [
                list(self._exportable_dict.keys()),
                list(self._exportable_dict.values()),
            ]
            return data[index.column()][index.row()]
        return None

    def setData(  # type: ignore # noqa: N802
        self,
        index: QModelIndex,
        value: Any,
        role: int,
    ) -> bool:
        """
        Set data at the given index for the given role.

        Args:
            index: a data index
            value: a value
            role: a role

        Returns:
            True if the data has been set, otherwise False
        """
        if self._exportable_dict is None or not index.isValid() or role != Qt.EditRole:
            return False
        existing_keys = list(self._exportable_dict.keys())
        key = existing_keys[index.row()]
        if index.column() == 0:
            if value in existing_keys:
                return False
            self._exportable_dict.swap_key(
                old=key,
                new=value,
            )
            self.dataChanged.emit(index, index)
            return True
        elif index.column() == 1:
            self._exportable_dict[key] = value
            self.dataChanged.emit(index, index)
            return True
        return False

    def rowCount(  # type: ignore # noqa: N802
        self,
        index: QModelIndex,
    ) -> int:
        """
        Get the number of rows/entries in the underlying exportable dict.

        Args:
            index: a data index for the parent data, if any

        Returns:
            The number of rows
        """
        return len(self._exportable_dict) if self._exportable_dict is not None else 0

    def columnCount(  # type: ignore # noqa: N802
        self,
        index: QModelIndex,
    ) -> int:
        """
        Get the number of columns in the underlying exportable dict, which is always 2 (key and value).

        Args:
            index: a data index for the parent data, if any

        Returns:
            The number of columns
        """
        return 2

    def flags(
        self,
        index: QModelIndex,
    ) -> Qt.ItemFlags:
        """
        Get the flags for the given index.

        Args:
            index: a data index

        Returns:
            The flags
        """
        if not index.isValid():
            return Qt.ItemIsEnabled

        return cast(Qt.ItemFlags, super().flags(index) | Qt.ItemIsEditable)

    def insertRows(  # type: ignore # noqa: N802
        self,
        row: int,
        count: int,
        parent: QModelIndex,
    ) -> bool:
        """
        Insert a number of rows before a given row.

        If row is 0, the rows are prepended to any existing rows in the parent.

        If row is rowCount(), the rows are appended to any existing rows in the parent.

        If parent has no children, a single column with count rows is inserted.

        Args:
            row: a row number
            count: a number of rows to be inserted
            parent: a parent index

        Returns:
            True if the rows where successfully inserted, otherwise False
        """
        self.beginInsertRows(
            parent,
            row,
            row + count - 1,
        )
        for i in range(count):
            key = self._get_next_new_available_key(
                base_index=i,
            )
            self._exportable_dict[key] = ''
        self.endInsertRows()
        return True

    def _get_next_new_available_key(
        self,
        base_index: int,
    ) -> str:
        n = 1
        while True:
            value = f'New entry {base_index + n}'
            if value not in self._exportable_dict.keys():
                return value
            n += 1

    def removeRows(  # type: ignore # noqa: N802
        self,
        row: int,
        count: int,
        parent: QModelIndex,
    ) -> bool:
        """
        Remove count rows starting with the given row under parent.

        Args:
            row: a row number
            count: a number of rows
            parent: a parent index

        Returns:
            True if the rows were successfully removed; otherwise False.
        """
        self.beginRemoveRows(
            parent,
            row,
            row + count - 1,
        )
        key = list(self._exportable_dict.keys())[row]
        self._exportable_dict = self._exportable_dict.__class__(
            {
                k: v for k, v in self._exportable_dict.items() if k != key
            },
        )
        self.endRemoveRows()
        return True



"""Models and functions used for listening to events."""
from abc import ABC
from dataclasses import dataclass
from typing import Generic, TypeVar


@dataclass
class Event(ABC):
    """A base event."""


E = TypeVar('E', bound=Event, contravariant=True)


class Listener(Generic[E], ABC):
    """A generic listener receiving events."""

    def on_event(
        self,
        event: E,
    ) -> None:
        """
        Receive an event.

        Args:
            event: an event
        """



"""Models and functions used for observing test of YesWeHack and tracker clients."""
from abc import ABC
from dataclasses import dataclass

from ywh2bt.core.configuration.root import RootConfiguration
from ywh2bt.core.configuration.tracker import TrackerConfiguration
from ywh2bt.core.configuration.yeswehack import YesWeHackConfiguration
from ywh2bt.core.listener import Event, Listener


@dataclass
class TesterEvent(Event, ABC):
    """A base event sent by a Tester."""


@dataclass
class TesterStartEvent(TesterEvent):
    """An event sent when a test starts."""

    configuration: RootConfiguration


@dataclass
class TesterEndEvent(TesterEvent):
    """An event sent when a test ends."""

    configuration: RootConfiguration


@dataclass
class TesterStartYesWeHackEvent(TesterEvent):
    """An event sent when a test of a YesWeJackConfiguration starts."""

    yeswehack_name: str
    yeswehack_configuration: YesWeHackConfiguration


@dataclass
class TesterEndYesWeHackEvent(TesterEvent):
    """An event sent when a test of a YesWeJackConfiguration ends."""

    yeswehack_name: str
    yeswehack_configuration: YesWeHackConfiguration


@dataclass
class TesterStartTrackerEvent(TesterEvent):
    """An event sent when a test of a Tracker starts."""

    tracker_name: str
    tracker_configuration: TrackerConfiguration


@dataclass
class TesterEndTrackerEvent(TesterEvent):
    """An event sent when a test of a Tracker ends."""

    tracker_name: str
    tracker_configuration: TrackerConfiguration


class TesterListener(Listener[TesterEvent], ABC):
    """A listener receiving events from a Tester."""


class NoOpTesterListener(TesterListener):
    """A listener receiving events from a Tester and doing nothing."""

    def on_event(
        self,
        event: TesterEvent,
    ) -> None:
        """
        Receive an event and do nothing.

        Args:
            event: an event
        """



"""Models for reports."""
from __future__ import annotations

from dataclasses import dataclass
from functools import partial
from types import MappingProxyType
from typing import (
    Any,
    Callable,
    Dict,
    List,
    Optional,
)

from typing_extensions import Protocol
from yeswehack.api import Report as YesWeHackRawApiReport

REPORT_PROPERTY_LABELS = MappingProxyType({
    'bug_type': 'Bug type',
    'scope': 'Scope',
    'end_point': 'Endpoint',
    'vulnerable_part': 'Vulnerable part',
    'part_name': 'Part name',
    'payload_sample': 'Payload',
    'application_finger_print': 'Application fingerprint',
})
REPORT_STATUS_TRANSLATIONS = MappingProxyType({
    'new': 'New',
    'under_review': 'Under Review',
    'reopen_under_review': 'Reopen Under Review',
    'more_info': 'Need More Info',
    'reopen_more_info': 'Reopen Need More Info',
    'accepted': 'Accepted',
    'ask_verif': 'Ask for fix verification',
    'wont_fix': "Won't fix",
    'informative': 'Informative',
    'resolved': 'Resolved',
    'rtfs': 'RTFS',
    'spam': 'Spam',
    'out_of_scope': 'Out Of Scope',
    'not_applicable': 'Not Applicable',
    'invalid': 'Invalid',
    'duplicate': 'Duplicate',
    'auto_close': 'Auto Close',
})


@dataclass
class Report:
    """A report."""

    raw_report: YesWeHackRawApiReport
    report_id: str
    title: Optional[str]
    local_id: str
    bug_type: BugType
    scope: str
    cvss: Cvss
    end_point: str
    vulnerable_part: str
    part_name: str
    payload_sample: str
    technical_environment: str
    description_html: str
    attachments: List[Attachment]
    hunter: Author
    logs: List[Log]
    status: str
    tracking_status: str
    program: ReportProgram
    priority: Optional[Priority] = None

    def get_last_tracking_status_update_log(
        self,
        tracker_name: str,
    ) -> Optional[Log]:
        """
        Get the log for the last tracking status update of a tracker, if any.

        Args:
            tracker_name: a tracker name

        Returns:
            a log
        """
        tracking_status_logs = list(filter(
            _all_log_filter(
                partial(
                    _log_type_equals,
                    log_type='tracking-status',
                ),
                partial(
                    _log_tracker_name_equals,
                    tracker_name=tracker_name,
                ),
                _not_log_filter(
                    partial(
                        _log_tracker_id_equals,
                        tracker_id=None,
                    ),
                ),
            ),
            self.logs,
        ))
        if tracking_status_logs:
            return tracking_status_logs[-1]
        return None

    def get_logs_after(
        self,
        log: Log,
    ) -> List[Log]:
        """
        Get all the logs that occurred after the given log.

        Args:
            log: a log

        Returns:
            the logs that occurred after the given log
        """
        log_index = self.logs.index(log)
        return self.logs[log_index + 1:]

    def get_comments_after(
        self,
        log: Log,
    ) -> List[Log]:
        """
        Get all the comments that occurred after the given log.

        Args:
            log: a log

        Returns:
            the comments that occurred after the given log
        """
        return list(filter(
            _log_is_public_comment,
            self.get_logs_after(
                log=log,
            ),
        ))

    def get_comments(
        self,
    ) -> List[Log]:
        """
        Get all the comments.

        Returns:
            the comments
        """
        return list(filter(
            _log_is_public_comment,
            self.logs,
        ))


@dataclass
class Attachment:
    """An attachment."""

    attachment_id: int
    name: str
    original_name: str
    mime_type: str
    size: int
    url: str
    data_loader: Callable[[], bytes]

    @property
    def data(self) -> bytes:
        """
        Get the data from the attachment.

        Returns:
            the data
        """
        return self.data_loader()


@dataclass
class Priority:
    """A priority."""

    name: str = ''


@dataclass
class ReportProgram:
    """Program details from a report."""

    title: str = ''
    slug: str = ''


@dataclass
class BugType:
    """A bug type."""

    name: str
    link: str
    remediation_link: str


@dataclass
class Cvss:
    """A CVSS."""

    criticity: str
    score: float
    vector: str


@dataclass
class Author:
    """An author."""

    username: str


@dataclass
class Log:
    """A log."""

    created_at: str
    log_id: int
    log_type: str
    private: bool
    author: Author
    message_html: str
    attachments: List[Attachment]


@dataclass
class CommentLog(Log):
    """A comment log."""


@dataclass
class CvssUpdateLog(Log):
    """A cvss-update log."""

    old_cvss: Cvss
    new_cvss: Cvss


@dataclass
class DetailsUpdateLog(Log):
    """A details-update log."""

    old_details: Optional[Dict[str, Any]]
    new_details: Optional[Dict[str, Any]]


@dataclass
class PriorityUpdateLog(Log):
    """A priority-update log."""

    new_priority: Optional[Priority]


@dataclass
class RewardLog(Log):
    """A reward log."""

    reward_type: str


@dataclass
class StatusUpdateLog(Log):
    """A status-update log."""

    old_status: Optional[Dict[str, Any]]
    new_status: Optional[Dict[str, Any]]


@dataclass
class CloseLog(Log):
    """A close log."""

    old_status: Optional[Dict[str, Any]]
    new_status: Optional[Dict[str, Any]]


@dataclass
class TrackingStatusLog(Log):
    """A tracking-status log."""

    tracker_name: Optional[str]
    tracker_url: Optional[str]
    tracker_id: Optional[str]


@dataclass
class TrackerUpdateLog(Log):
    """A tracker-update log."""

    tracker_name: Optional[str]
    tracker_url: Optional[str]
    tracker_id: Optional[str]
    tracker_token: Optional[str]


@dataclass
class TrackerMessageLog(Log):
    """A tracker-message log."""

    tracker_name: Optional[str]
    tracker_url: Optional[str]
    tracker_id: Optional[str]


@dataclass
class FixVerifiedLog(Log):
    """A fix-verified log."""

    verified: bool


class _LogFilterProtocol(Protocol):

    def __call__(
        self,
        log: Log,
    ) -> bool:
        ...  # noqa: WPS428


def _apply_log_filter(
    log: Log,
    filter_fn: _LogFilterProtocol,
) -> bool:
    return filter_fn(
        log=log,
    )


def _all_log_filter(
    *filters: _LogFilterProtocol,
) -> _LogFilterProtocol:
    def fn(  # noqa: WPS430
        log: Log,
    ) -> bool:
        return all(
            map(
                partial(
                    _apply_log_filter,
                    log,
                ),
                filters,
            ),
        )

    return fn


def _not_log_filter(
    filter_fn: _LogFilterProtocol,
) -> _LogFilterProtocol:
    def fn(  # noqa: WPS430
        log: Log,
    ) -> bool:
        return not _apply_log_filter(
            log=log,
            filter_fn=filter_fn,
        )

    return fn


def _log_type_equals(
    log_type: str,
    log: Log,
) -> bool:
    return log.log_type == log_type


def _log_private_equals(
    private: bool,
    log: Log,
) -> bool:
    return log.private == private


def _log_tracker_name_equals(
    tracker_name: Optional[str],
    log: TrackingStatusLog,
) -> bool:
    return log.tracker_name == tracker_name


def _log_tracker_id_equals(
    tracker_id: Optional[str],
    log: TrackingStatusLog,
) -> bool:
    return log.tracker_id == tracker_id


_log_is_comment = partial(
    _log_type_equals,
    log_type='comment',
)
_log_is_public = partial(
    _log_private_equals,
    private=False,
)
_log_is_public_comment = _all_log_filter(
    _log_is_comment,
    _log_is_public,
)



"""Models and functions used for data synchronisation between YesWeHack and ServiceNow trackers."""
from typing import (
    Any,
    Dict,
    Union,
)

from aiosnow import (
    Client,
    Selector,
    TableModel,
)
from aiosnow.models import BaseTableModel
from aiosnow.models._schema import fields  # noqa: WPS436
from aiosnow.models.attachment.schema import AttachmentModelSchema
from aiosnow.query import Condition
from aiosnow.request import (
    Response,
    methods,
)

Record = Dict[str, Any]


class UserModelSchema:
    """User API Schema."""

    sys_id = fields.String(is_primary=True)
    user_name = fields.String()


class UserModel(TableModel, UserModelSchema):
    """User API Model."""


class InMemoryAttachmentModel(BaseTableModel, AttachmentModelSchema):
    """Attachment API Model."""

    _client: Client

    async def create(self, payload: Record) -> Response:
        """
        Create a new record.

        Args:
            payload: New record payload

        Raises:
            AttributeError: always because this method is not supported on this object
        """
        raise AttributeError(
            "Attachment doesn't support create(), use upload() instead",
        )

    async def update(self, selection: Union[Condition, str], payload: Record) -> Response:
        """
        Update matching record.

        Args:
            selection: Condition or ID of object to update
            payload: Update payload

        Raises:
            AttributeError: always because this method is not supported on this object
        """
        raise AttributeError(
            "Attachment doesn't support update()",
        )

    @property
    def _api_url(self) -> str:
        return f'{self._client.base_url}/api/now/attachment'

    async def download(
        self,
        selection: Union[Selector, Condition, str],
    ) -> bytes:
        """
        Download file.

        Args:
            selection: Attachment selection

        Returns:
            The content of the attachment
        """
        meta = await self.get_one(selection)
        data = await self.request(
            methods.GET,
            url=meta['download_link'],
            resolve=False,
            decode=False,
        )
        return await data.read()

    async def upload(
        self,
        table_name: str,
        record_sys_id: str,
        file_name: str,
        content_type: str,
        content: bytes,
    ) -> Response:
        """
        Upload a file.

        Args:
            table_name: Table name, e.g. incident
            record_sys_id: Sys id of the record to attach to
            file_name: Source file name
            content_type: Content type
            content: Content

        Returns:
            ClientResponse
        """
        return await self.request(
            methods.POST,
            url=f'{self._api_url}/file',
            params={
                'table_name': table_name,
                'table_sys_id': record_sys_id,
                'file_name': file_name,
            },
            headers={
                'Content-type': content_type,
            },
            payload=content,
        )



"""Models and functions used in data typing."""
from typing import Any, Union, cast

from PySide2.QtCore import Signal, SignalInstance


def as_signal_instance(
    signal: Union[Signal, Any],
) -> SignalInstance:
    """
    Cast a Signal object to a SignalInstance.

    PySide2 magic / static typing / mypy.

    Args:
        signal: a signal object

    Returns:
        The signal instance
    """
    return cast(SignalInstance, signal)



"""Models and functions used for line-edits containing sensible data."""
from __future__ import annotations

from functools import partial
from typing import Any, Optional

from PySide2.QtCore import Signal
from PySide2.QtGui import QIcon
from PySide2.QtWidgets import (
    QHBoxLayout,
    QLineEdit,
    QPushButton,
    QWidget,
)

from ywh2bt.gui.widgets import constants
from ywh2bt.gui.widgets.typing import as_signal_instance


class SecretLineEditWidget(QWidget):
    """A widget consisting of a line-edit in password echo mode and a button temporarily enables normal echo mode."""

    text_edited: Signal = Signal(str)

    _icon_show: QIcon
    _icon_hide: QIcon

    _line_edit: QLineEdit
    _button: QPushButton

    def __init__(
        self,
        parent: Optional[QWidget] = None,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """
        Initialize self.

        Args:
            parent: a parent widget
            args: extra arguments
            kwargs: extra keyword arguments
        """
        super().__init__(
            parent,
            *args,
            **kwargs,
        )
        self._icon_show = QIcon(':/resources/icons/show.png')
        self._icon_hide = QIcon(':/resources/icons/hide.png')

        self._init_ui()

    def _init_ui(self) -> None:
        layout = QHBoxLayout(self)

        self._line_edit = self._create_line_edit()

        self._button = self._create_button()

        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(self._line_edit)
        layout.addWidget(self._button)

    def _create_line_edit(
        self,
    ) -> QLineEdit:
        widget = QLineEdit(self)
        widget.setEchoMode(QLineEdit.Password)
        as_signal_instance(widget.textEdited[str]).connect(
            as_signal_instance(self.text_edited).emit,
        )
        return widget

    def _create_button(
        self,
    ) -> QPushButton:
        widget = QPushButton(self)
        widget.setIcon(self._icon_hide)
        widget.setToolTip('Reveal secret')
        widget.setFixedWidth(constants.SMALL_BUTTON_WIDTH)
        widget.setIconSize(constants.SMALL_BUTTON_ICON_SIZE)
        as_signal_instance(widget.pressed).connect(
            partial(
                self._update_widget,
                QLineEdit.Normal,
                self._icon_show,
            ),
        )
        as_signal_instance(widget.released).connect(
            partial(
                self._update_widget,
                QLineEdit.Password,
                self._icon_hide,
            ),
        )
        return widget

    def _update_widget(
        self,
        echo_mode: QLineEdit.EchoMode,
        icon: QIcon,
    ) -> None:
        self._line_edit.setEchoMode(echo_mode)
        self._button.setIcon(icon)

    def setText(  # noqa: N802
        self,
        text: str,
    ) -> None:
        """
        Set the text of the line-edit.

        Args:
            text: a text
        """
        self._line_edit.setText(text)



"""Models and classes related to exportable objects."""
from abc import ABC, abstractmethod
from typing import Generic, TypeVar

T = TypeVar('T')


class Exportable(Generic[T], ABC):
    """A class representing an object that can be exported."""

    @abstractmethod
    def export(self) -> T:
        """Export the object."""



"""Models and functions used for data loading/saving."""
from abc import ABC, abstractmethod
from typing import Any, Generic, TextIO, TypeVar

from ywh2bt.core.configuration.error import BaseAttributeError
from ywh2bt.core.configuration.root import RootConfiguration
from ywh2bt.core.exceptions import CoreException
from ywh2bt.core.serde import SerDe, SerDeError

LoadType = TypeVar('LoadType')
DumpType = TypeVar('DumpType')


class LoaderError(CoreException):
    """A loading/saving error."""


class _BaseLoader(ABC, Generic[LoadType, DumpType]):
    @abstractmethod
    def load(
        self,
        stream: TextIO,
    ) -> LoadType:
        """
        Load data from a stream.

        Args:
            stream: a stream
        """

    @abstractmethod
    def save(
        self,
        data: DumpType,
        stream: TextIO,
    ) -> None:
        """
        Save data in a stream.

        Args:
            data: data to be saved
            stream: a stream
        """


class DataLoader(_BaseLoader[Any, Any]):
    """Generic data loader."""

    _serde: SerDe

    def __init__(
        self,
        serde: SerDe,
    ) -> None:
        """
        Initialize self.

        Args:
            serde: serializer/deserializer used when loading/saving
        """
        self._serde = serde

    def load(
        self,
        stream: TextIO,
    ) -> Any:
        """
        Load data from a stream.

        Args:
            stream: a stream

        Returns:
            the loaded data

        Raises:
            LoaderError: if the data couldn't be deserialized
        """
        try:
            data = self._serde.deserialize_from_stream(
                stream=stream,
            )
        except SerDeError as e:
            raise LoaderError('Unable to deserialize data') from e
        return data

    def save(
        self,
        data: Any,
        stream: TextIO,
    ) -> None:
        """
        Save data in a stream.

        Args:
            data: data to be saved
            stream: a stream

        Raises:
            LoaderError: if the data couldn't be serialized
        """
        try:
            self._serde.serialize_to_stream(
                data=data,
                stream=stream,
            )
        except SerDeError as e:
            raise LoaderError('Unable to serialize data') from e


class RootConfigurationLoader(_BaseLoader[RootConfiguration, RootConfiguration]):
    """Root configuration data loader."""

    _loader: DataLoader

    def __init__(
        self,
        serde: SerDe,
    ) -> None:
        """
        Initialize self.

        Args:
            serde: serializer/deserializer used when loading/saving
        """
        self._loader = DataLoader(
            serde=serde,
        )

    def load(
        self,
        stream: TextIO,
    ) -> RootConfiguration:
        """
        Load configuration from a stream.

        Args:
            stream: a stream

        Returns:
            the loaded configuration

        Raises:
            LoaderError: if the configuration couldn't be loaded
        """
        data = self._loader.load(
            stream=stream,
        )
        if not isinstance(data, dict):
            data_type = type(data)
            raise LoaderError(f'Expecting {dict} for root configuration ; got {data_type}')
        try:
            return RootConfiguration(**data)
        except BaseAttributeError as e:
            raise LoaderError('Unable to load root configuration') from e

    def save(
        self,
        data: RootConfiguration,
        stream: TextIO,
    ) -> None:
        """
        Save configuration in a stream.

        Args:
            data: data to be saved
            stream: a stream
        """
        self._loader.save(
            data=data.export(),
            stream=stream,
        )



"""Models and functions used for main GUI window."""
from __future__ import annotations

import sys
from typing import List, Optional, Union, cast

from PySide2.QtCore import QByteArray, QEvent, QSettings, QSize, Qt, Signal
from PySide2.QtGui import QIcon, QKeySequence
from PySide2.QtWidgets import (
    QAction,
    QMainWindow,
    QMenu,
    QMessageBox,
    QSizePolicy,
    QStyle,
    QToolBar,
    QWidget,
)

from ywh2bt.core.core import AVAILABLE_FORMATS
from ywh2bt.gui.dialog.file import FileFormatDialogFilters
from ywh2bt.gui.dialog.schema import SchemaDocumentationDialog
from ywh2bt.gui.widgets.logs_widget import LogEntry
from ywh2bt.gui.widgets.root_configuration_entry import RootConfigurationEntry
from ywh2bt.gui.widgets.root_configurations_widget import RootConfigurationsWidget
from ywh2bt.gui.widgets.typing import as_signal_instance
from ywh2bt.version import __VERSION__

__UI_VERSION__ = 1

_MENU_SEPARATOR = '-'
_TOOL_BAR_SEPARATOR = '-'
_TOOL_BAR_SPACER = '<->'


class MainWindow(QMainWindow):  # noqa: WPS214
    """Main window."""

    _log_entry_available: Signal = Signal(LogEntry)

    _file_format_dialog_filters: FileFormatDialogFilters
    _settings: QSettings
    _schema_doc_dialog: Optional[SchemaDocumentationDialog]

    _action_new: QAction
    _action_open: QAction
    _action_save: QAction
    _action_save_as: QAction
    _action_reload: QAction
    _action_test: QAction
    _action_sync: QAction
    _action_exit: QAction
    _action_schema_doc: QAction
    _action_about: QAction

    _root_configurations_widget: RootConfigurationsWidget

    def __init__(
        self,
        settings: QSettings,
    ) -> None:
        """
        Initialize self.

        Args:
            settings: application settings
        """
        super().__init__()
        self._file_format_dialog_filters = FileFormatDialogFilters(
            formats=AVAILABLE_FORMATS,
        )
        self._settings = settings
        self._schema_doc_dialog = None
        self._init_ui()
        self._restore_settings()

    def _init_ui(self) -> None:
        self._create_actions()
        self._create_menus()

        tool_bar = self._create_main_tool_bar()
        self.addToolBar(
            Qt.LeftToolBarArea,
            tool_bar,
        )

        self._root_configurations_widget = self._create_root_configurations_widget()

        self.setCentralWidget(self._root_configurations_widget)

        self.statusBar().show()

    def _create_actions(
        self,
    ) -> None:
        self._action_new = self._create_new_action()
        self._action_open = self._create_open_action()
        self._action_save = self._create_save_action()
        self._action_save_as = self._create_save_as_action()
        self._action_reload = self._create_reload_action()
        self._action_test = self._create_test_action()
        self._action_sync = self._create_sync_action()
        self._action_exit = self._create_exit_action()
        self._action_schema_doc = self._create_schema_doc_action()
        self._action_about = self._create_about_action()

    def _create_menus(
        self,
    ) -> None:
        bar = self.menuBar()

        file_menu = bar.addMenu('&File')
        file_menu.setObjectName('menu_file')
        _add_menu_items(
            menu=file_menu,
            items=[
                self._action_new,
                self._action_open,
                self._action_save,
                self._action_save_as,
                self._action_reload,
                _MENU_SEPARATOR,
                self._action_exit,
            ],
        )

        run_menu = bar.addMenu('&Run')
        run_menu.setObjectName('menu_run')
        _add_menu_items(
            menu=run_menu,
            items=[
                self._action_test,
                self._action_sync,
            ],
        )

        help_menu = bar.addMenu('&Help')
        help_menu.setObjectName('menu_help')
        _add_menu_items(
            menu=help_menu,
            items=[
                self._action_schema_doc,
                self._action_about,
            ],
        )

    def _create_new_action(
        self,
    ) -> QAction:
        action = QAction('&New', self)
        action.setObjectName('action_new')
        action.setIcon(self.style().standardIcon(QStyle.SP_FileIcon))
        action.setShortcuts(QKeySequence.New)
        action.setStatusTip('Create a new configuration file')
        as_signal_instance(action.triggered).connect(
            self._on_new_file_triggered,
        )
        return action

    def _create_open_action(
        self,
    ) -> QAction:
        action = QAction('&Open...', self)
        action.setObjectName('action_open')
        action.setIcon(self.style().standardIcon(QStyle.SP_DialogOpenButton))
        action.setShortcuts(QKeySequence.Open)
        action.setStatusTip('Open a configuration file')
        as_signal_instance(action.triggered).connect(
            self._on_open_file_triggered,
        )
        return action

    def _create_save_action(
        self,
    ) -> QAction:
        action = QAction('&Save...', self)
        action.setObjectName('action_save')
        action.setEnabled(False)
        action.setIcon(self.style().standardIcon(QStyle.SP_DialogSaveButton))
        action.setShortcuts(QKeySequence.Save)
        action.setStatusTip('Save the current configuration file')
        as_signal_instance(action.triggered).connect(
            self._on_save_configuration_triggered,
        )
        return action

    def _create_save_as_action(
        self,
    ) -> QAction:
        action = QAction('&Save as...', self)
        action.setObjectName('action_save_as')
        action.setEnabled(False)
        action.setStatusTip('Save the current configuration in a new file')
        as_signal_instance(action.triggered).connect(
            self._on_save_as_configuration_triggered,
        )
        return action

    def _create_reload_action(
        self,
    ) -> QAction:
        action = QAction('&Reload from disk', self)
        action.setObjectName('action_reload')
        action.setEnabled(False)
        action.setIcon(self.style().standardIcon(QStyle.SP_BrowserReload))
        action.setShortcuts(QKeySequence.Refresh)
        action.setStatusTip('Reload the current configuration file from the disk')
        as_signal_instance(action.triggered).connect(
            self._on_reload_configuration_triggered,
        )
        return action

    def _create_test_action(
        self,
    ) -> QAction:
        action = QAction('&Test...', self)
        action.setObjectName('action_test')
        action.setEnabled(False)
        action.setIcon(self.style().standardIcon(QStyle.SP_DialogApplyButton))
        action.setStatusTip('Test the connection to the trackers and to YesWeHack programs')
        as_signal_instance(action.triggered).connect(
            self._on_test_configuration_triggered,
        )
        return action

    def _create_sync_action(
        self,
    ) -> QAction:
        action = QAction('&Sync...', self)
        action.setObjectName('action_sync')
        action.setEnabled(False)
        action.setIcon(self.style().standardIcon(QStyle.SP_MediaPlay))
        action.setStatusTip('Synchronize the current configuration file')
        as_signal_instance(action.triggered).connect(
            self._on_sync_configuration_triggered,
        )
        return action

    def _create_about_action(
        self,
    ) -> QAction:
        action = QAction('&About', self)
        action.setObjectName('action_about')
        action.setIcon(QIcon(':/resources/icons/ywh2bt.png'))
        action.setStatusTip('Show information about this application')
        as_signal_instance(action.triggered).connect(
            self._on_about_triggered,
        )
        return action

    def _create_schema_doc_action(
        self,
    ) -> QAction:
        action = QAction('&Schema documentation', self)
        action.setObjectName('action_schema_doc')
        action.setIcon(self.style().standardIcon(QStyle.SP_DialogHelpButton))
        action.setStatusTip('Show configuration schema documentation')
        as_signal_instance(action.triggered).connect(
            self._on_schema_doc_triggered,
        )
        return action

    def _create_exit_action(
        self,
    ) -> QAction:
        action = QAction('E&xit', self)
        action.setObjectName('action_exit')
        action.setShortcuts(QKeySequence.Quit)
        action.setStatusTip('Exit the application')
        as_signal_instance(action.triggered).connect(
            self._on_exit_triggered,
        )
        return action

    def _create_main_tool_bar(
        self,
    ) -> QWidget:
        tool_bar = self.addToolBar('Main tools')
        tool_bar.setObjectName('tools_tool_bar')

        _add_tool_bar_items(
            tool_bar=tool_bar,
            items=[
                self._action_new,
                self._action_open,
                self._action_save,
                self._action_reload,
                _TOOL_BAR_SEPARATOR,
                self._action_test,
                self._action_sync,
                _TOOL_BAR_SPACER,
                _TOOL_BAR_SEPARATOR,
                self._action_schema_doc,
                self._action_about,
            ],
        )

        return tool_bar

    def _create_empty_widget(
        self,
    ) -> QWidget:
        toolbar = QToolBar(self)
        toolbar.setToolButtonStyle(Qt.ToolButtonTextUnderIcon)

        spacer_begin = QWidget(toolbar)
        spacer_begin.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Preferred)
        spacer_end = QWidget(toolbar)
        spacer_end.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Preferred)

        toolbar.addWidget(spacer_begin)
        toolbar.addAction(self._action_new)
        toolbar.addAction(self._action_open)
        toolbar.addWidget(spacer_end)

        return toolbar

    def _create_root_configurations_widget(
        self,
    ) -> RootConfigurationsWidget:
        widget = RootConfigurationsWidget(
            parent=self,
            settings=self._settings,
            empty_widget=self._create_empty_widget(),
        )
        as_signal_instance(widget.entry_modified).connect(
            self._on_entry_modified,
        )
        as_signal_instance(widget.current_entry_changed).connect(
            self._on_current_entry_changed,
        )
        as_signal_instance(widget.entry_saved).connect(
            self._on_entry_saved,
        )
        return widget

    def _on_entry_modified(
        self,
        entry: RootConfigurationEntry,
    ) -> None:
        self._update_actions(
            entry=entry,
        )

    def _on_entry_saved(
        self,
        entry: RootConfigurationEntry,
    ) -> None:
        self.statusBar().showMessage(
            f'{entry.name} saved',
        )

    def _on_current_entry_changed(
        self,
        entry: Optional[RootConfigurationEntry],
    ) -> None:
        self._update_actions(
            entry=entry,
        )

    def _update_actions(
        self,
        entry: Optional[RootConfigurationEntry],
    ) -> None:
        self._update_save_action(
            entry=entry,
        )
        self._update_save_as_action(
            entry=entry,
        )
        self._update_reload_action(
            entry=entry,
        )
        self._update_test_action(
            entry=entry,
        )
        self._update_sync_action(
            entry=entry,
        )

    def _update_save_action(
        self,
        entry: Optional[RootConfigurationEntry],
    ) -> None:
        enabled = False
        if entry and not entry.is_empty() and entry.has_changed():
            enabled = True
        self._action_save.setEnabled(enabled)

    def _update_save_as_action(
        self,
        entry: Optional[RootConfigurationEntry],
    ) -> None:
        enabled = False
        if entry and not entry.is_empty():
            enabled = True
        self._action_save_as.setEnabled(enabled)

    def _update_reload_action(
        self,
        entry: Optional[RootConfigurationEntry],
    ) -> None:
        enabled = False
        if entry and entry.file:
            enabled = True
        self._action_reload.setEnabled(enabled)

    def _update_test_action(
        self,
        entry: Optional[RootConfigurationEntry],
    ) -> None:
        enabled = False
        if entry and entry.configuration and entry.configuration.is_valid():
            enabled = True
        self._action_test.setEnabled(enabled)

    def _update_sync_action(
        self,
        entry: Optional[RootConfigurationEntry],
    ) -> None:
        enabled = False
        if entry and entry.configuration and entry.configuration.is_valid():
            enabled = True
        self._action_sync.setEnabled(enabled)

    def _on_new_file_triggered(
        self,
    ) -> None:
        self._root_configurations_widget.create_new_configuration()

    def _on_open_file_triggered(
        self,
    ) -> None:
        self._root_configurations_widget.prompt_open_configuration()

    def _on_save_configuration_triggered(
        self,
    ) -> None:
        self._root_configurations_widget.save_current_entry()

    def _on_save_as_configuration_triggered(
        self,
    ) -> None:
        self._root_configurations_widget.save_as_current_entry()

    def _on_reload_configuration_triggered(
        self,
    ) -> None:
        self._root_configurations_widget.reload_current_configuration()

    def _on_test_configuration_triggered(
        self,
    ) -> None:
        self._root_configurations_widget.test_current_entry()

    def _on_sync_configuration_triggered(
        self,
    ) -> None:
        self._root_configurations_widget.synchronize_current_entry()

    def _on_exit_triggered(
        self,
    ) -> None:
        self.close()

    def _on_about_triggered(
        self,
    ) -> None:
        python_version = '.'.join(map(str, sys.version_info[:3]))
        about_text = [
            f'<b>Core version</b>: {__VERSION__}',
            f'<b>UI version</b>: {__UI_VERSION__}',
            f'Python: {python_version}',
        ]
        QMessageBox.about(
            self,
            'ywh2bt-gui',
            '<br/>\n'.join(about_text),
        )

    def _on_schema_doc_triggered(
        self,
    ) -> None:
        if not self._schema_doc_dialog:
            self._schema_doc_dialog = SchemaDocumentationDialog(
                self,
            )
            self._schema_doc_dialog.resize(
                QSize(600, 600),
            )
            self._restore_schema_doc_geometry_from_settings()
        self._schema_doc_dialog.show()

    def closeEvent(  # noqa: N802
        self,
        event: QEvent,
    ) -> None:
        """
        Handle 'close' event.

        Args:
            event: a close event
        """
        if self._root_configurations_widget.has_changed_entries():
            reply = QMessageBox.question(
                self,
                'Exit',
                'Are you sure you want to exit?\nUnsaved changes might be lost.',
                QMessageBox.Yes | QMessageBox.No,
                QMessageBox.No,
            )
            if reply == QMessageBox.No:
                event.ignore()
                return
        self._save_settings()
        event.accept()
        super().closeEvent(event)

    def _save_settings(
        self,
    ) -> None:
        self._settings.setValue('main_geometry', self.saveGeometry().toBase64())
        self._settings.setValue('main_state', self.saveState(__UI_VERSION__).toBase64())
        if self._schema_doc_dialog:
            self._settings.setValue('schema_doc_geometry', self._schema_doc_dialog.saveGeometry().toBase64())

    def _restore_settings(
        self,
    ) -> None:
        self._restore_geometry_from_settings()
        self._restore_state_from_settings()

    def _restore_geometry_from_settings(
        self,
    ) -> None:
        if not self._settings.contains('main_geometry'):
            return
        geometry = QByteArray.fromBase64(cast(QByteArray, self._settings.value('main_geometry')))
        if geometry:
            self.restoreGeometry(geometry)

    def _restore_state_from_settings(
        self,
    ) -> None:
        if not self._settings.contains('main_state'):
            return
        state = QByteArray.fromBase64(cast(QByteArray, self._settings.value('main_state')))
        if state:
            self.restoreState(state, __UI_VERSION__)

    def _restore_schema_doc_geometry_from_settings(
        self,
    ) -> None:
        if not self._schema_doc_dialog or not self._settings.contains('schema_doc_geometry'):
            return
        geometry = QByteArray.fromBase64(cast(QByteArray, self._settings.value('schema_doc_geometry')))
        if geometry:
            self._schema_doc_dialog.restoreGeometry(geometry)


def _add_menu_items(
    menu: QMenu,
    items: List[Union[str, QAction]],
) -> None:
    for item in items:
        if item == _MENU_SEPARATOR:
            menu.addSeparator()
        elif isinstance(item, QAction):
            menu.addAction(item)


def _add_tool_bar_items(
    tool_bar: QToolBar,
    items: List[Union[str, QAction]],
) -> None:
    for item in items:
        if item == _TOOL_BAR_SEPARATOR:
            tool_bar.addSeparator()
        elif item == _TOOL_BAR_SPACER:
            spacer = QWidget(tool_bar)
            spacer.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
            tool_bar.addWidget(spacer)
        elif isinstance(item, QAction):
            tool_bar.addAction(item)



"""Models and functions used for YesWeHack data mapping."""
from dataclasses import dataclass
from typing import Any, Dict, List, Union

from yeswehack.api import (  # noqa: N811
    Attachment as YesWeHackRawApiAttachment,
    Author as YesWeHackRawApiAuthor,
    BugType as YesWeHackRawApiBugType,
    CVSS as YesWeHackRawApiCvss,
    Log as YesWeHackRawApiLog,
    Priority as YesWeHackRawApiPriority,
    Report as YesWeHackRawApiReport,
)

from ywh2bt.core.api.models.report import (  # noqa: WPS235
    Attachment,
    Author,
    BugType,
    CloseLog,
    CommentLog,
    Cvss,
    CvssUpdateLog,
    DetailsUpdateLog,
    FixVerifiedLog,
    Log,
    Priority,
    PriorityUpdateLog,
    Report,
    ReportProgram,
    RewardLog,
    StatusUpdateLog,
    TrackerMessageLog,
    TrackerUpdateLog,
    TrackingStatusLog,
)
from ywh2bt.core.html import (
    cleanup_attachments_and_urls_from_html,
    cleanup_ywh_redirects_from_html,
    cleanup_ywh_redirects_from_text,
)


@dataclass
class MappingContext:
    """A mapping context."""

    yeswehack_domain: str


def map_raw_report(
    context: MappingContext,
    raw_report: YesWeHackRawApiReport,
) -> Report:
    """
    Map a raw API report to a local report.

    Args:
        context: a mapping context
        raw_report: a raw report

    Returns:
        a local report
    """
    attachments = _map_raw_attachments(
        context=context,
        raw_attachments=raw_report.attachments,
    )
    description_html = cleanup_ywh_redirects_from_html(
        ywh_domain=context.yeswehack_domain,
        html=raw_report.description_html,
    )
    description_html, attachments = cleanup_attachments_and_urls_from_html(
        html=description_html,
        attachments=attachments,
    )
    bug_type = _map_raw_bug_type(
        context=context,
        raw_bug_type=raw_report.bug_type,
    )
    cvss = _map_raw_cvss(
        context=context,
        raw_cvss=raw_report.cvss,
    )
    priority = _map_raw_priority(
        context=context,
        raw_priority=raw_report.priority,
    ) if raw_report.priority else None
    hunter = _map_raw_author(
        context=context,
        raw_author=raw_report.hunter,
    )
    status = _map_raw_status(
        context=context,
        raw_status=raw_report.status,
    )
    logs = map_raw_logs(
        context=context,
        raw_logs=raw_report.logs or [],
    )
    return Report(
        raw_report=raw_report,
        report_id=str(raw_report.id),
        title=raw_report.title,
        local_id=raw_report.local_id,
        bug_type=bug_type,
        scope=raw_report.scope,
        cvss=cvss,
        end_point=raw_report.end_point,
        vulnerable_part=raw_report.vulnerable_part,
        part_name=raw_report.part_name,
        payload_sample=raw_report.payload_sample,
        technical_environment=raw_report.technical_environment,
        description_html=description_html,
        attachments=attachments,
        hunter=hunter,
        status=status,
        tracking_status=raw_report.tracking_status,
        logs=logs,
        priority=priority,
        program=_map_raw_report_program(
            context=context,
            raw_program=raw_report.program or {},
        ),
    )


def _map_raw_bug_type(
    context: MappingContext,
    raw_bug_type: YesWeHackRawApiBugType,
) -> BugType:
    return BugType(
        name=raw_bug_type.name,
        link=raw_bug_type.link,
        remediation_link=raw_bug_type.remediation_link,
    )


def _map_raw_priority(
    context: MappingContext,
    raw_priority: YesWeHackRawApiPriority,
) -> Priority:
    return Priority(
        name=raw_priority.name,
    )


def _map_raw_report_program(
    context: MappingContext,
    raw_program: Dict[Any, Any],
) -> ReportProgram:
    return ReportProgram(
        title=raw_program.get('title', ''),
        slug=raw_program.get('slug', ''),
    )


def _map_raw_cvss(
    context: MappingContext,
    raw_cvss: YesWeHackRawApiCvss,
) -> Cvss:
    return Cvss(
        criticity=raw_cvss.criticity,
        score=raw_cvss.score,
        vector=raw_cvss.vector,
    )


def _map_raw_attachments(
    context: MappingContext,
    raw_attachments: List[YesWeHackRawApiAttachment],
) -> List[Attachment]:
    return [
        map_raw_attachment(
            context=context,
            raw_attachment=raw_attachment,
        )
        for raw_attachment in raw_attachments
    ]


def map_raw_attachment(
    context: MappingContext,
    raw_attachment: YesWeHackRawApiAttachment,
) -> Attachment:
    """
    Map a raw API attachment to a local attachment.

    Args:
        context: a mapping context
        raw_attachment: a raw attachment

    Returns:
        a local attachment
    """
    return Attachment(
        attachment_id=raw_attachment.id,
        name=raw_attachment.name,
        original_name=raw_attachment.original_name,
        mime_type=raw_attachment.mime_type,
        size=raw_attachment.size,
        url=raw_attachment.url,
        data_loader=lambda: raw_attachment.data,
    )


def _map_raw_author(
    context: MappingContext,
    raw_author: Union[YesWeHackRawApiAuthor, Dict[str, Any]],
) -> Author:
    default_username = 'Anonymous'
    if isinstance(raw_author, YesWeHackRawApiAuthor):
        username = raw_author.username or default_username
    else:
        username = raw_author.get('username', default_username)
    return Author(
        username=username,
    )


def _map_raw_status(
    context: MappingContext,
    raw_status: Dict[str, Any],
) -> str:
    return raw_status.get('workflow_state', '')


def map_raw_logs(
    context: MappingContext,
    raw_logs: List[YesWeHackRawApiLog],
) -> List[Log]:
    """
    Map a list of raw API logs to a list of local logs.

    Args:
        context: a mapping context
        raw_logs: a list of raw logs

    Returns:
        a list of local logs
    """
    return [
        map_raw_log(
            context=context,
            raw_log=raw_log,
        )
        for raw_log in raw_logs
    ]


def map_raw_log(  # noqa: WPS210,WPS212,WPS231
    context: MappingContext,
    raw_log: YesWeHackRawApiLog,
) -> Log:
    """
    Map a raw API log to a local log.

    Args:
        context: a mapping context
        raw_log: a raw log

    Returns:
        a local log
    """
    created_at = raw_log.created_at
    log_id = raw_log.id
    log_type = raw_log.type
    private = raw_log.private
    author = _map_raw_author(
        context=context,
        raw_author=raw_log.author,
    ) if raw_log.author else Author(
        username='Anonymous',
    )
    attachments = _map_raw_attachments(
        context=context,
        raw_attachments=raw_log.attachments,
    )
    message_html = cleanup_ywh_redirects_from_html(
        ywh_domain=context.yeswehack_domain,
        html=raw_log.message_html or '',
    )
    message_html, attachments = cleanup_attachments_and_urls_from_html(
        html=message_html,
        attachments=attachments,
    )
    if raw_log.type == 'close':
        return CloseLog(
            created_at=created_at,
            log_id=log_id,
            log_type=log_type,
            private=private,
            author=author,
            message_html=message_html,
            attachments=attachments,
            old_status=raw_log.old_status,
            new_status=raw_log.status,
        )
    if raw_log.type == 'comment' or (raw_log.type == 'assign' and message_html):
        return CommentLog(
            created_at=created_at,
            log_id=log_id,
            log_type=log_type,
            private=private,
            author=author,
            message_html=message_html,
            attachments=attachments,
        )
    if raw_log.type == 'cvss-update':
        return CvssUpdateLog(
            created_at=created_at,
            log_id=log_id,
            log_type=log_type,
            private=private,
            author=author,
            message_html=message_html,
            attachments=attachments,
            old_cvss=_map_raw_cvss(
                context=context,
                raw_cvss=raw_log.old_cvss,
            ),
            new_cvss=_map_raw_cvss(
                context=context,
                raw_cvss=raw_log.new_cvss,
            ),
        )
    if raw_log.type == 'details-update':
        return DetailsUpdateLog(
            created_at=created_at,
            log_id=log_id,
            log_type=log_type,
            private=private,
            author=author,
            message_html=message_html,
            attachments=attachments,
            old_details=raw_log.old_details,
            new_details=raw_log.new_details,
        )
    if raw_log.type == 'priority-update':
        return PriorityUpdateLog(
            created_at=created_at,
            log_id=log_id,
            log_type=log_type,
            private=private,
            author=author,
            message_html=message_html,
            attachments=attachments,
            new_priority=_map_raw_priority(
                context=context,
                raw_priority=raw_log.priority,
            ) if raw_log.priority else None,
        )
    if raw_log.type == 'reward':
        return RewardLog(
            created_at=created_at,
            log_id=log_id,
            log_type=log_type,
            private=private,
            author=author,
            message_html=message_html,
            attachments=attachments,
            reward_type=raw_log.reward_type,
        )
    if raw_log.type == 'status-update':
        return StatusUpdateLog(
            created_at=created_at,
            log_id=log_id,
            log_type=log_type,
            private=private,
            author=author,
            message_html=message_html,
            attachments=attachments,
            old_status=raw_log.old_status,
            new_status=raw_log.status,
        )
    if raw_log.type == 'tracking-status':
        return TrackingStatusLog(
            created_at=created_at,
            log_id=log_id,
            log_type=log_type,
            private=private,
            author=author,
            message_html=message_html,
            attachments=attachments,
            tracker_name=raw_log.tracker_name,
            tracker_url=cleanup_ywh_redirects_from_text(
                ywh_domain=context.yeswehack_domain,
                text=raw_log.tracker_url,
            ) if raw_log.tracker_url else None,
            tracker_id=raw_log.tracker_id,
        )
    if raw_log.type == 'tracker-update':
        return TrackerUpdateLog(
            created_at=created_at,
            log_id=log_id,
            log_type=log_type,
            private=private,
            author=author,
            message_html=message_html,
            attachments=attachments,
            tracker_name=raw_log.tracker_name,
            tracker_url=cleanup_ywh_redirects_from_text(
                ywh_domain=context.yeswehack_domain,
                text=raw_log.tracker_url,
            ) if raw_log.tracker_url else None,
            tracker_id=raw_log.tracker_id,
            tracker_token=raw_log.tracker_token,
        )
    if raw_log.type == 'tracker-message':
        return TrackerMessageLog(
            created_at=created_at,
            log_id=log_id,
            log_type=log_type,
            private=private,
            author=author,
            message_html=message_html,
            attachments=attachments,
            tracker_name=raw_log.tracker_name,
            tracker_url=cleanup_ywh_redirects_from_text(
                ywh_domain=context.yeswehack_domain,
                text=raw_log.tracker_url,
            ) if raw_log.tracker_url else None,
            tracker_id=raw_log.tracker_id,
        )
    if raw_log.type == 'fix-verified':
        return FixVerifiedLog(
            created_at=created_at,
            log_id=log_id,
            log_type=log_type,
            private=private,
            author=author,
            message_html=message_html,
            attachments=attachments,
            verified=raw_log.fix_verified,
        )
    return Log(
        created_at=created_at,
        log_id=log_id,
        log_type=log_type,
        private=private,
        author=author,
        message_html=message_html,
        attachments=attachments,
    )



"""Models and functions used for check box with hint."""
from typing import Any, Dict, Optional

from PySide2.QtCore import Qt, Signal
from PySide2.QtWidgets import QCheckBox, QHBoxLayout, QLabel, QWidget

from ywh2bt.gui.widgets.typing import as_signal_instance


class HintedCheckBoxWidget(QWidget):
    """A widget consisting of a checkbox with a hint."""

    stateChanged: Signal = Signal(Qt.CheckState)  # noqa: WPS115, N815

    _check_box: QCheckBox
    _label: QLabel

    _state_hints: Dict[Qt.CheckState, str]

    def __init__(
        self,
        parent: Optional[QWidget] = None,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """
        Initialize self.

        Args:
            parent: a parent widget
            args: extra arguments
            kwargs: extra keyword arguments
        """
        super().__init__(
            parent,
            *args,
            **kwargs,
        )
        self._state_hints = {}
        self._init_ui()

    def _init_ui(self) -> None:
        layout = QHBoxLayout(self)

        self._check_box = self._create_check_box()

        self._label = self._create_label()

        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(self._check_box)
        layout.addWidget(self._label)
        layout.addStretch(1)

    def _create_check_box(
        self,
    ) -> QCheckBox:
        widget = QCheckBox(self)
        as_signal_instance(widget.stateChanged).connect(
            self._on_check_box_state_changed,
        )
        return widget

    def _on_check_box_state_changed(
        self,
        state: int,
    ) -> None:
        states = {
            0: Qt.Unchecked,
            1: Qt.PartiallyChecked,
            2: Qt.Checked,
        }
        self._update_label_for_state(
            state=states[state],
        )
        as_signal_instance(self.stateChanged).emit(states[state])

    def set_state_hint(
        self,
        state: Qt.CheckState,
        hint: str,
    ) -> None:
        """
        Set a hint for a CheckState.

        Args:
            state: a state
            hint: a hint
        """
        self._state_hints[state] = hint

    def _update_label_for_state(
        self,
        state: Qt.CheckState,
    ) -> None:
        self._label.setText(self._state_hints.get(state, ''))

    def _create_label(
        self,
    ) -> QLabel:
        return QLabel(self)

    def setTristate(  # noqa: N802
        self,
        tristate: bool,
    ) -> None:
        """
        Set whether the checkbox is a tri-state checkbox.

        Args:
            tristate: a flag indicating if the checkbox is a tri-state checkbox
        """
        self._check_box.setTristate(tristate)

    def setCheckState(  # noqa: N802
        self,
        state: Qt.CheckState,
    ) -> None:
        """
        Set the checkbox's check state.

        Args:
            state: a state
        """
        self._check_box.setCheckState(state)



"""Version."""
import os

import pkg_resources
import tomlkit

try:
    __VERSION__ = pkg_resources.get_distribution('ywh2bt').version
except pkg_resources.DistributionNotFound:
    # fallback if project is not installed
    pyproject_path = os.path.join(
        os.path.dirname(__file__),
        '..',
        'pyproject.toml',
    )
    with open(pyproject_path) as pyproject_file:
        pyproject = pyproject_file.read()
    __VERSION__ = 'Unknown'
    tool = tomlkit.api.parse(pyproject).get('tool')
    if tool:
        poetry = tool.get('poetry')
        if poetry:
            __VERSION__ = poetry.get('version', 'Unknown')



"""Models and functions used for formatting reports data."""

from abc import ABC, abstractmethod
from string import Template
from typing import Optional

from singledispatchmethod import singledispatchmethod
from typing_extensions import Protocol

from ywh2bt.core.api.models.report import (
    CloseLog,
    CommentLog,
    CvssUpdateLog,
    DetailsUpdateLog,
    FixVerifiedLog,
    Log,
    PriorityUpdateLog,
    REPORT_PROPERTY_LABELS,
    REPORT_STATUS_TRANSLATIONS,
    Report,
    RewardLog,
    StatusUpdateLog,
)


class _ValueTransformer(Protocol):
    def __call__(
        self,
        value: str,
    ) -> str:
        ...  # noqa: WPS428


def _identity_transformer(
    value: str,
) -> str:
    return value


class ReportMessageFormatter(ABC):
    """A base report formatter."""

    _report_title_template: Template
    _report_description_template: Template
    _close_log_template: Template
    _comment_body_template: Template
    _comment_log_template: Template
    _cvss_update_log_template: Template
    _status_update_log_template: Template
    _fix_verified_log_template: Template
    _details_update_log_template: Template
    _details_update_log_line_template: Template
    _priority_update_log_template: Template
    _reward_log_template: Template
    _value_transformer: _ValueTransformer

    def __init__(  # noqa: WPS211
        self,
        report_title_template: Template,
        report_description_template: Template,
        close_log_template: Template,
        comment_body_template: Template,
        comment_log_template: Template,
        cvss_update_log_template: Template,
        status_update_log_template: Template,
        fix_verified_log_template: Template,
        details_update_log_template: Template,
        details_update_log_line_template: Template,
        priority_update_log_template: Template,
        reward_log_template: Template,
        value_transformer: Optional[_ValueTransformer] = None,
    ):
        """
        Initialize self.

        Args:
            report_title_template: a template for a report title
            report_description_template: a template for a report description
            close_log_template: a template for a CloseLog
            comment_body_template: a template for an issue comment
            comment_log_template: a template for a template for a CommentLog
            cvss_update_log_template: a template for a CvssUpdateLog
            status_update_log_template: a template for a StatusUpdateLog
            fix_verified_log_template: a template for a FixVerifiedLog
            details_update_log_template: a template for a DetailsUpdateLog
            details_update_log_line_template: a template for entries of a DetailsUpdateLog
            priority_update_log_template: a template for a PriorityUpdateLog
            reward_log_template: a template for a RewardLog
            value_transformer: a transformer for values
        """
        self._report_title_template = report_title_template
        self._report_description_template = report_description_template
        self._close_log_template = close_log_template
        self._comment_body_template = comment_body_template
        self._comment_log_template = comment_log_template
        self._cvss_update_log_template = cvss_update_log_template
        self._status_update_log_template = status_update_log_template
        self._fix_verified_log_template = fix_verified_log_template
        self._details_update_log_template = details_update_log_template
        self._details_update_log_line_template = details_update_log_line_template
        self._priority_update_log_template = priority_update_log_template
        self._reward_log_template = reward_log_template
        self._value_transformer = value_transformer or _identity_transformer

    def _transform_value(
        self,
        value: str,
    ) -> str:
        return self._value_transformer(
            value=value,
        )

    def format_report_title(
        self,
        report: Report,
    ) -> str:
        """
        Get a formatted title for a report.

        Args:
            report: a report

        Returns:
            a formatted title
        """
        return self._report_title_template.substitute(
            local_id=report.local_id,
            title=self._transform_value(
                value=report.title or '',
            ),
        )

    def format_report_description(
        self,
        report: Report,
    ) -> str:
        """
        Get a formatted description for a report.

        Args:
            report: a report

        Returns:
            a formatted description
        """
        return self._report_description_template.substitute(
            local_id=report.local_id,
            title=self._transform_value(
                value=report.title or '',
            ),
            priority_name=report.priority.name if report.priority else '',
            bug_type_label=self._get_property_label('bug_type'),
            bug_type_name=report.bug_type.name,
            bug_type_link=report.bug_type.link,
            bug_type_remediation_link=report.bug_type.remediation_link or '/',
            scope_label=self._get_property_label('scope'),
            scope=self._transform_value(
                value=report.scope,
            ),
            cvss_criticity=report.cvss.criticity,
            cvss_score=report.cvss.score,
            cvss_vector=report.cvss.vector,
            end_point_label=self._get_property_label('end_point'),
            end_point=self._transform_value(
                value=report.end_point,
            ),
            vulnerable_part_label=self._get_property_label('vulnerable_part'),
            vulnerable_part=self._transform_value(
                value=report.vulnerable_part,
            ),
            part_name_label=self._get_property_label('part_name'),
            part_name=self._transform_value(
                value=report.part_name,
            ),
            payload_sample_label=self._get_property_label('payload_sample'),
            payload_sample=self._transform_value(
                value=report.payload_sample or '',
            ),
            technical_environment=self._transform_value(
                value=report.technical_environment or '',
            ),
            description=self.transform_report_description_html(
                description_html=report.description_html,
            ),
        )

    def _get_property_label(
        self,
        property_name: str,
    ) -> str:
        return REPORT_PROPERTY_LABELS.get(property_name, property_name)

    @abstractmethod
    def transform_report_description_html(
        self,
        description_html: str,
    ) -> str:
        """
        Transform the report description.

        Args:
            description_html: a report description in HTML
        """

    def format_log(
        self,
        log: Log,
    ) -> str:
        """
        Get a formatted comment for a log.

        Args:
            log: a log

        Returns:
            a formatted log
        """
        return self._comment_body_template.substitute(
            log_id=log.log_id,
            author=log.author.username,
            date=log.created_at,
            body=self._transform_log(
                log,
            ),
        )

    @abstractmethod
    def transform_html(
        self,
        html: str,
    ) -> str:
        """
        Transform html.

        Args:
            html: some HTML
        """

    @singledispatchmethod
    def _transform_log(
        self,
        log: Log,
    ) -> str:
        return self.transform_html(
            html=log.message_html,
        )

    @_transform_log.register
    def _transform_comment_log(
        self,
        log: CommentLog,
    ) -> str:
        return self._comment_log_template.substitute(
            comment=self.transform_html(
                html=log.message_html,
            ),
        )

    @_transform_log.register
    def _transform_cvss_update_log(
        self,
        log: CvssUpdateLog,
    ) -> str:
        return self._cvss_update_log_template.substitute(
            old_cvss_criticity=log.old_cvss.criticity,
            old_cvss_score=log.old_cvss.score,
            new_cvss_criticity=log.new_cvss.criticity,
            new_cvss_score=log.new_cvss.score,
        )

    @_transform_log.register
    def _transform_status_update_log(
        self,
        log: StatusUpdateLog,
    ) -> str:
        return self._status_update_log_template.substitute(
            old_status=self._translate_status(
                status=(log.old_status or {}).get('workflow_state') or '',
            ),
            new_status=self._translate_status(
                status=(log.new_status or {}).get('workflow_state') or '',
            ),
            comment=self.transform_html(
                html=log.message_html,
            ),
        )

    @_transform_log.register
    def _transform_fix_verified_log(
        self,
        log: FixVerifiedLog,
    ) -> str:
        return self._fix_verified_log_template.substitute(
            verified='Fix verified' if log.verified else 'Fix refused',
            comment=self.transform_html(
                html=log.message_html,
            ),
        )

    @_transform_log.register
    def _transform_close_log(
        self,
        log: CloseLog,
    ) -> str:
        return self._close_log_template.substitute(
            old_status=self._translate_status(
                status=(log.old_status or {}).get('workflow_state') or '',
            ),
            new_status=self._translate_status(
                status=(log.new_status or {}).get('workflow_state') or '',
            ),
            comment=self.transform_html(
                html=log.message_html,
            ),
        )

    def _translate_status(
        self,
        status: str,
    ) -> str:
        return REPORT_STATUS_TRANSLATIONS.get(status, '')

    @_transform_log.register
    def _transform_details_update_log(
        self,
        log: DetailsUpdateLog,
    ) -> str:
        details_lines = []
        for updated_property, new_value in (log.new_details or {}).items():
            old_value = (log.old_details or {}).get(updated_property, '')
            details_lines.append(
                self._details_update_log_line_template.substitute(
                    updated_property=self._get_property_label(
                        property_name=updated_property,
                    ),
                    old_value=self._transform_value(
                        value=old_value or '',
                    ),
                    new_value=self._transform_value(
                        value=new_value or '',
                    ),
                ),
            )
        return self._details_update_log_template.substitute(
            details_lines=''.join(details_lines),
        )

    @_transform_log.register
    def _transform_priority_update_log(
        self,
        log: PriorityUpdateLog,
    ) -> str:
        return self._priority_update_log_template.substitute(
            new_priority=log.new_priority.name if log.new_priority else 'Undefined',
        )

    @_transform_log.register
    def _format_reward_log(
        self,
        log: RewardLog,
    ) -> str:
        return self._reward_log_template.substitute(
            comment=self.transform_html(
                html=log.message_html,
            ),
        )



"""Models and functions used for GitHub attachments upload."""
from json.decoder import JSONDecodeError
from typing import Any, Dict, Optional, cast
from urllib.parse import urlsplit

import requests
from bs4 import BeautifulSoup, Tag  # type: ignore
from github.Issue import Issue
from requests_toolbelt import MultipartEncoder  # type: ignore

from ywh2bt.core.api.models.report import Attachment
from ywh2bt.core.configuration.trackers.github import GitHubConfiguration


class GitHubAttachmentUploader:
    """An attachment uploader."""

    _configuration: GitHubConfiguration
    _github_domain: str

    _login_form_field_names = (
        'authenticity_token',
        'commit',
        'timestamp_secret',
        'webauthn-support',
        'webauthn-iuvpaa-support',
        'required_field_59db',
        'utf8',
    )

    def __init__(
        self,
        configuration: GitHubConfiguration,
    ):
        """
        Initialize self.

        Args:
            configuration: a GitHub configuration
        """
        self._configuration = configuration
        domain = 'github.com'
        if configuration.url != 'https://api.github.com':
            domain = urlsplit(cast(str, configuration.url)).netloc
        self._github_domain = domain

    def upload_attachment(
        self,
        issue: Issue,
        attachment: Attachment,
    ) -> Optional[str]:
        """
        Upload an attachment to an issue.

        Args:
            issue: an issue
            attachment: an attachment

        Returns:
            the url or the uploaded attachment
        """
        session = self._login()
        if not session:
            return None

        upload_policies_assets_authenticity_token = self._get_upload_policies_assets_authenticity_token(
            session=session,
            issue=issue,
        )
        if not upload_policies_assets_authenticity_token:
            return None

        policies_assets_response_data = self._upload_policies_assets(
            session=session,
            issue=issue,
            attachment=attachment,
            authenticity_token=upload_policies_assets_authenticity_token,
        )
        if not policies_assets_response_data:
            return None

        upload_response = self._upload(
            session=session,
            attachment=attachment,
            url=policies_assets_response_data['upload_url'],
            form=policies_assets_response_data['form'],
        )
        if not upload_response:
            return None

        url = self._upload_asset(
            session=session,
            issue=issue,
            url=policies_assets_response_data['asset_upload_url'],
            authenticity_token=policies_assets_response_data['asset_upload_authenticity_token'],
        )

        return cast(
            str,
            url or policies_assets_response_data['asset']['href'],
        )

    def _get_upload_policies_assets_authenticity_token(
        self,
        session: requests.Session,
        issue: Issue,
    ) -> Optional[str]:
        issue_response = self._github_request(
            session=session,
            method='GET',
            path=f'/{issue.repository.full_name}/issues/{issue.number}',
        )
        if not issue_response:
            return None

        return self._extract_upload_policies_asset_authenticity_token(
            html=issue_response.text,
        )

    def _extract_upload_policies_asset_authenticity_token(
        self,
        html: str,
    ) -> Optional[str]:
        file_attachment_tag = BeautifulSoup(
            html,
            features='lxml',
        ).find(
            'file-attachment',
        )

        if not file_attachment_tag:
            return None

        authenticity_token = file_attachment_tag.get('data-upload-policy-authenticity-token', None)
        if authenticity_token is None:
            csrf_policy_input = file_attachment_tag.find(
                'input', {
                    'class': 'js-data-upload-policy-url-csrf',
                },
            )
            if csrf_policy_input is not None:
                authenticity_token = csrf_policy_input.get('value', None)

        return cast(str, authenticity_token)

    def _upload_policies_assets(
        self,
        session: requests.Session,
        issue: Issue,
        attachment: Attachment,
        authenticity_token: str,
    ) -> Optional[Dict[str, Any]]:
        data = MultipartEncoder(
            fields={
                'name': attachment.original_name.split('/')[-1],
                'size': str(len(attachment.data)),
                'content_type': attachment.mime_type,
                'authenticity_token': authenticity_token,
                'repository_id': str(issue.repository.id),
            },
        )
        response = self._github_request(
            session=session,
            method='POST',
            path='/upload/policies/assets',
            data=data,
            headers={
                'Content-Type': data.content_type,
                'Content-Length': str(data.len),
            },
        )
        if not response or response.status_code != 201:
            return None
        response_data = response.json()
        if 'asset' not in response_data:
            return None
        return cast(Dict[str, Any], response_data)

    def _upload(
        self,
        session: requests.Session,
        attachment: Attachment,
        url: str,
        form: Dict[str, Any],
    ) -> bool:
        upload_data = MultipartEncoder(
            fields={
                **form,
                'file': (attachment.original_name.split('/')[-1], attachment.data, attachment.mime_type),
            },
        )
        upload_response = _url_request(
            session=session,
            method='POST',
            url=url,
            data=upload_data,
            headers={
                'Content-Type': upload_data.content_type,
                'Content-Length': str(upload_data.len),
            },
        )
        if not upload_response:  # noqa: WPS531
            return False
        return True

    def _upload_asset(
        self,
        session: requests.Session,
        issue: Issue,
        url: str,
        authenticity_token: str,
    ) -> Optional[str]:
        asset_upload_data = MultipartEncoder(
            fields={
                'authenticity_token': authenticity_token,
            },
        )
        repository_url = f'https://{self._github_domain}/{issue.repository.full_name}'
        asset_upload_response = self._github_request(
            session=session,
            method='PUT',
            path=url,
            data=asset_upload_data,
            headers={
                'Content-Type': asset_upload_data.content_type,
                'Content-Length': str(asset_upload_data.len),
                'Referer': f'{repository_url}/issues/{issue.number}',
                'Accept': 'application/json',
            },
        )
        if not asset_upload_response:
            return None
        try:
            asset_upload_response_data = asset_upload_response.json()
        except JSONDecodeError:
            return None
        return cast(Optional[str], asset_upload_response_data.get('href', None))

    def _login(self) -> Optional[requests.Session]:
        session = requests.Session()
        session.verify = self._configuration.verify
        login_page = self._github_request(
            session=session,
            method='GET',
            path='/login',
        )
        if not login_page:
            return None
        form = BeautifulSoup(login_page.text, features='lxml').find('form')
        fields = self._extract_login_fields(
            form=form,
        )
        session_response = self._github_request(
            session=session,
            method='POST',
            path='/session',
            data={
                **fields,
                'login': self._configuration.login,
                'password': self._configuration.password,
            },
        )
        if session_response and session_response.status_code == 200:
            return session
        return None

    def _extract_login_fields(
        self,
        form: Tag,
    ) -> Dict[str, Any]:
        fields: Dict[str, Any] = {}
        for field_name in self._login_form_field_names:
            value = form.find(
                'input', {
                    'name': field_name,
                },
            )
            if value:
                fields[field_name] = value['value']
        return fields

    def _github_request(
        self,
        session: requests.Session,
        method: str,
        path: str,
        data: Optional[Any] = None,
        headers: Optional[Dict[str, Any]] = None,
    ) -> Optional[requests.Response]:
        return _url_request(
            session=session,
            method=method,
            url=f'https://{self._github_domain}{path}',
            data=data,
            headers=headers,
        )


def _url_request(
    session: requests.Session,
    method: str,
    url: str,
    data: Optional[Any] = None,
    headers: Optional[Dict[str, Any]] = None,
) -> Optional[requests.Response]:
    try:
        return session.request(
            method=method,
            url=url,
            data=data,
            headers=headers,
        )
    except requests.RequestException:
        return None



"""Models and functions used for data synchronisation between YesWeHack and GitLab trackers."""
from __future__ import annotations

import os
import re
from copy import deepcopy
from datetime import datetime
from typing import (
    Dict,
    Iterable,
    List,
    Optional,
    Tuple,
    cast,
)

import requests
from gitlab import (  # type: ignore
    Gitlab,
    GitlabError,
)
from gitlab.v4.objects import (  # type: ignore
    Project,
    ProjectIssue,
    ProjectIssueNote,
)

from ywh2bt.core.api.formatter.markdown import ReportMessageMarkdownFormatter
from ywh2bt.core.api.models.report import (
    Attachment,
    Log,
    Report,
)
from ywh2bt.core.api.tracker import (
    SendLogsResult,
    TrackerAttachment,
    TrackerClient,
    TrackerClientError,
    TrackerIssue,
    TrackerIssueComment,
    TrackerIssueComments,
)
from ywh2bt.core.configuration.trackers.gitlab import GitLabConfiguration

_RE_IMAGE = re.compile(pattern=r'!\[([^\]]+)]\(([^)]+)\)')
_RE_CONTENT_DISPOSITION_FILENAME = re.compile(pattern='filename="([^"]+)";?')
_TEXT_MAX_SIZE = 1000000


class GitLabTrackerClientError(TrackerClientError):
    """A GitLab tracker client error."""


class GitLabTrackerClient(TrackerClient[GitLabConfiguration]):
    """A GitLab tracker client."""

    _session: requests.Session
    _gitlab: Gitlab
    _message_formatter: ReportMessageMarkdownFormatter
    _default_author_name: str

    def __init__(
        self,
        configuration: GitLabConfiguration,
    ):
        """
        Initialize self.

        Args:
            configuration: a GitLab configuration
        """
        super().__init__(
            configuration=configuration,
        )
        self._default_author_name = 'Anonymous'
        self._session = requests.Session()
        self._session.verify = configuration.verify
        self._gitlab = Gitlab(
            url=configuration.url,
            private_token=configuration.token,
            session=self._session,
        )
        self._message_formatter = ReportMessageMarkdownFormatter()

    @property
    def tracker_type(self) -> str:
        """
        Get the type of the tracker client.

        Returns:
            the type of the  tracker client
        """
        return 'GitLab'

    def _build_tracker_issue(
        self,
        issue_id: str,
        issue_url: str,
        closed: bool,
    ) -> TrackerIssue:
        return TrackerIssue(
            tracker_url=cast(str, self.configuration.url),
            project=cast(str, self.configuration.project),
            issue_id=issue_id,
            issue_url=issue_url,
            closed=closed,
        )

    def get_tracker_issue(
        self,
        issue_id: str,
    ) -> Optional[TrackerIssue]:
        """
        Get a tracker issue.

        Args:
            issue_id: an issue id

        Returns:
            The issue if it exists, else None
        """
        try:
            gitlab_issue = self._get_gitlab_issue(
                issue_id=issue_id,
            )
        except GitLabTrackerClientError:
            return None
        if gitlab_issue is None:
            return None
        return self._build_tracker_issue(
            issue_id=issue_id,
            issue_url=gitlab_issue.web_url,
            closed=gitlab_issue.state == 'closed',
        )

    def get_tracker_issue_comments(
        self,
        issue_id: str,
        exclude_comments: Optional[List[str]] = None,
    ) -> TrackerIssueComments:
        """
        Get a list of comments on an issue.

        Args:
            issue_id: an issue id
            exclude_comments: an optional list of comment to exclude

        Returns:
            The list of comments
        """
        try:
            gitlab_issue = self._get_gitlab_issue(
                issue_id=issue_id,
            )
        except GitLabTrackerClientError:
            return []
        if not gitlab_issue:
            return []
        return self._extract_comments(
            gitlab_issue=gitlab_issue,
            exclude_comments=exclude_comments,
        )

    def send_report(
        self,
        report: Report,
    ) -> TrackerIssue:
        """
        Send a report to the tracker.

        Args:
            report: a report

        Returns:
            information about the sent report
        """
        self._ensure_auth()
        gitlab_project = self._get_gitlab_project()
        title = self._message_formatter.format_report_title(
            report=report,
        )
        description = self._message_formatter.format_report_description(
            report=report,
        ) + self._get_attachments_list_description(
            attachments=report.attachments,
        )
        external_description = ''
        description_attachment = None
        if len(description) > _TEXT_MAX_SIZE:
            external_description = description
            description_attachment = self._build_external_description_attachment(
                name=f'report-{report.local_id.replace("#", "")}-description.md',
            )
            report_copy = deepcopy(report)
            report_copy.description_html = (
                '<p>This report description is too large to fit into a GitLab issue. '
                + f'See attachment <a href="{description_attachment.url}">{description_attachment.original_name}</a> '
                + 'for more details.</p>'
            )
            description = self._message_formatter.format_report_description(
                report=report_copy,
            ) + self._get_attachments_list_description(
                attachments=[
                    description_attachment,
                    *report.attachments,
                ],
            )
        description, external_description = self._replace_attachments_references(
            uploads=self._upload_attachments(
                gitlab_project=gitlab_project,
                attachments=report.attachments,
            ),
            referencing_texts=[
                description,
                external_description,
            ],
        )
        if description_attachment:
            description_attachment.data_loader = lambda: bytes(external_description, 'utf-8')
            description = self._replace_attachments_references(
                uploads=self._upload_attachments(
                    gitlab_project=gitlab_project,
                    attachments=[
                        description_attachment,
                    ],
                ),
                referencing_texts=[
                    description,
                ],
            )[0]
        gitlab_issue = self._create_issue(
            gitlab_project=gitlab_project,
            title=title,
            description=description,
            confidential=self.configuration.confidential or False,
        )
        return self._build_tracker_issue(
            issue_id=str(gitlab_issue.id),
            issue_url=gitlab_issue.web_url,
            closed=False,
        )

    def _build_external_description_attachment(
        self,
        name: str,
    ) -> Attachment:
        return Attachment(
            attachment_id=0,
            name=name,
            original_name=name,
            mime_type='text/markdown',
            size=0,
            url=f'http://tracker/external/{name}',
            data_loader=lambda: bytes('', 'utf-8'),
        )

    def _extract_comments(
        self,
        gitlab_issue: ProjectIssue,
        exclude_comments: Optional[List[str]] = None,
    ) -> List[TrackerIssueComment]:
        return [
            self._extract_comment(
                gitlab_note=gitlab_note,
            )
            for gitlab_note in reversed(gitlab_issue.notes.list())
            if exclude_comments is None or str(gitlab_note.id) not in exclude_comments
        ]

    def _extract_comment(
        self,
        gitlab_note: ProjectIssueNote,
    ) -> TrackerIssueComment:
        gitlab_body = gitlab_note.body
        inline_images = _RE_IMAGE.findall(gitlab_body)
        comment_attachments: Dict[str, TrackerAttachment] = {}
        for _, inline_path in inline_images:
            attachment = self._download_attachment(
                path=inline_path,
            )
            if attachment:
                comment_attachments[inline_path] = attachment
        return TrackerIssueComment(
            created_at=self._parse_date(
                date=gitlab_note.created_at,
            ),
            author=gitlab_note.author.get('name', self._default_author_name),
            comment_id=str(gitlab_note.id),
            body=gitlab_body,
            attachments=comment_attachments,
        )

    def _download_attachment(
        self,
        path: str,
    ) -> Optional[TrackerAttachment]:
        url = f'{self.configuration.url}/{self.configuration.project}/{path}'
        try:
            response = requests.get(url)
        except requests.RequestException:
            return None
        if not response.ok:
            return None
        content_disposition = response.headers.get('Content-Disposition')
        filename = None
        if content_disposition:
            match = _RE_CONTENT_DISPOSITION_FILENAME.search(content_disposition)
            if match:
                filename = match.group(1)
        if not content_disposition or not filename:
            filename = os.path.basename(path)
        return TrackerAttachment(
            filename=filename,
            mime_type=response.headers.get('Content-Type', 'text/plain'),
            content=response.content,
        )

    def send_logs(
        self,
        tracker_issue: TrackerIssue,
        logs: Iterable[Log],
    ) -> SendLogsResult:
        """
        Send logs to the tracker.

        Args:
            tracker_issue: information about the tracker issue
            logs: a list of comments

        Raises:
            GitLabTrackerClientError: if an error occurs

        Returns:
            information about the sent comments
        """
        self._ensure_auth()
        tracker_comments = SendLogsResult(
            tracker_issue=tracker_issue,
            added_comments=[],
        )
        gitlab_project = self._get_gitlab_project()
        gitlab_issue = self._get_gitlab_issue(
            issue_id=tracker_issue.issue_id,
        )
        if not gitlab_issue:
            raise GitLabTrackerClientError(
                f'GitLab issue {tracker_issue.issue_id} not found in project {self.configuration.project}',
            )
        for log in logs:
            gitlab_comment = self._add_comment(
                gitlab_project=gitlab_project,
                gitlab_issue=gitlab_issue,
                log=log,
            )
            tracker_comments.added_comments.append(
                TrackerIssueComment(
                    created_at=self._parse_date(
                        date=gitlab_comment.created_at,
                    ),
                    author=gitlab_comment.author.get('name', self._default_author_name),
                    comment_id=str(gitlab_comment.id),
                    body=gitlab_comment.body,
                    attachments={},
                ),
            )
        return tracker_comments

    def _parse_date(
        self,
        date: str,
    ) -> datetime:
        return datetime.strptime(
            date,
            '%Y-%m-%dT%H:%M:%S.%f%z',
        )

    def test(
        self,
    ) -> None:
        """
        Test the client.

        Raises:
            GitLabTrackerClientError: if the test failed
        """
        try:
            self._gitlab.auth()
        except GitlabError as e:
            raise GitLabTrackerClientError('Unable to log in with GitLab API client') from e

    def _get_gitlab_project(
        self,
    ) -> Project:
        try:
            return self._gitlab.projects.get(
                self.configuration.project,
            )
        except GitlabError as e:
            raise GitLabTrackerClientError(
                f'Unable to get GitLab project {self.configuration.project}',
            ) from e

    def _get_gitlab_issue(
        self,
        issue_id: str,
    ) -> Optional[ProjectIssue]:
        gitlab_project = self._get_gitlab_project()
        try:
            gitlab_issues = gitlab_project.issues.list(all=False, as_list=False)
        except GitlabError as e:
            raise GitLabTrackerClientError(
                f'Unable to get GitLab issues for project {self.configuration.project}',
            ) from e
        issue_id_int = int(issue_id)
        for gitlab_issue in gitlab_issues:
            if gitlab_issue.id == issue_id_int:
                return gitlab_issue
        return None

    def _replace_attachments_references(
        self,
        uploads: List[Tuple[Attachment, str]],
        referencing_texts: List[str],
    ) -> List[str]:
        for attachment, upload_url in uploads:
            referencing_texts = [
                text.replace(
                    attachment.url,
                    f'{self.configuration.url}/{self.configuration.project}{upload_url}',
                )
                for text in referencing_texts
            ]
        return referencing_texts

    def _upload_attachments(
        self,
        gitlab_project: Project,
        attachments: List[Attachment],
    ) -> List[Tuple[Attachment, str]]:
        try:
            return [
                (
                    attachment,
                    gitlab_project.upload(attachment.original_name, attachment.data)['url'],
                )
                for attachment in attachments
            ]
        except GitlabError as e:
            raise GitLabTrackerClientError(
                f'Unable to upload attachments for project {self.configuration.project} to GitLab',
            ) from e

    def _get_attachments_list_description(
        self,
        attachments: List[Attachment],
    ) -> str:
        attachments_lines = []
        if attachments:
            attachments_lines = [
                '',
                '**Attachments**:',
            ]
            for attachment in attachments:
                attachments_lines.append(
                    f'- [{attachment.original_name}]({attachment.url})',
                )
            attachments_lines.append('')
        return '\n'.join(attachments_lines)

    def _create_issue(
        self,
        gitlab_project: Project,
        title: str,
        description: str,
        confidential: bool,
    ) -> ProjectIssue:
        issue_data = {
            'title': title,
            'description': description,
            'confidential': confidential,
        }
        try:
            return gitlab_project.issues.create(issue_data)
        except GitlabError as e:
            raise GitLabTrackerClientError(
                f'Unable to create issue for project {self.configuration.project} to GitLab',
            ) from e

    def _add_comment(
        self,
        gitlab_project: Project,
        gitlab_issue: ProjectIssue,
        log: Log,
    ) -> ProjectIssueNote:
        comment_body = self._message_formatter.format_log(
            log=log,
        ) + self._get_attachments_list_description(
            attachments=log.attachments,
        )
        external_body = ''
        body_attachment = None
        if len(comment_body) > _TEXT_MAX_SIZE:
            external_body = comment_body
            body_attachment = self._build_external_description_attachment(
                name=f'comment-{log.log_id}-description.md',
            )
            log_copy = deepcopy(log)
            log_copy.message_html = (
                '<p>This comment is too large to fit into a GitLab comment. '
                + f'See attachment <a href="{body_attachment.url}">{body_attachment.original_name}</a> '
                + 'for more details.</p>'
            )
            comment_body = self._message_formatter.format_log(
                log=log_copy,
            ) + self._get_attachments_list_description(
                attachments=[
                    body_attachment,
                    *log.attachments,
                ],
            )
        comment_body, external_body = self._replace_attachments_references(
            uploads=self._upload_attachments(
                gitlab_project=gitlab_project,
                attachments=log.attachments,
            ),
            referencing_texts=[
                comment_body,
                external_body,
            ],
        )
        if body_attachment:
            body_attachment.data_loader = lambda: bytes(external_body, 'utf-8')
            comment_body = self._replace_attachments_references(
                uploads=self._upload_attachments(
                    gitlab_project=gitlab_project,
                    attachments=[
                        body_attachment,
                    ],
                ),
                referencing_texts=[
                    comment_body,
                ],
            )[0]
        try:
            return gitlab_issue.notes.create({
                'body': comment_body,
            })
        except GitlabError as e:
            raise GitLabTrackerClientError(
                f'Unable to add GitLab comment for issue {gitlab_issue} in project {self.configuration.project}',
            ) from e

    def _ensure_auth(
        self,
    ) -> None:
        try:
            self._gitlab.auth()
        except GitlabError as e:
            raise GitLabTrackerClientError('Unable to authenticate to GitLab') from e



"""Models and functions used for attributes container dict GUI."""
import re
from dataclasses import dataclass
from functools import partial
from string import Template
from typing import Any, Dict, List, Optional, Tuple, Type, cast

from PySide2.QtCore import QFile, QSize, Qt, Signal
from PySide2.QtGui import QIcon
from PySide2.QtWidgets import (
    QHBoxLayout,
    QLayout,
    QTabWidget,
    QToolButton,
    QWidget,
)

from ywh2bt.core.configuration.attribute import AttributesContainer, AttributesContainerDict
from ywh2bt.core.configuration.subtypable import SubtypableMetaclass
from ywh2bt.gui.widgets import constants
from ywh2bt.gui.widgets.attribute.attributes_container_dict_entry import AttributesContainerDictEntry
from ywh2bt.gui.widgets.typing import as_signal_instance

T_AC = AttributesContainer
T_ACD = AttributesContainerDict[T_AC]

ICON_PATH_TEMPLATES: Tuple[Template, ...] = (
    Template(':/resources/icons/types/${type_name}-2x.png'),
    Template(':/resources/icons/types/${type_name}.png'),
)


@dataclass
class _ButtonDescription:
    type_name: str
    icon_type_name: str
    entry_args: Dict[str, Any]


class AttributesContainerDictWidget(QWidget):
    """Attributes container dict GUI."""

    container_dict_changed: Signal = Signal(AttributesContainerDict)
    _container_dict_class: Type[T_ACD]
    _container_dict: Optional[T_ACD]

    _empty_widget: QWidget
    _tab_widget: QTabWidget

    def __init__(
        self,
        container_dict_class: Type[T_ACD],
        container_dict: Optional[T_ACD] = None,
        parent: Optional[QWidget] = None,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """
        Initialize self.

        Args:
            container_dict_class: a concrete class extending AttributesContainerDict
            container_dict: an attributes container dict
            parent: a parent widget
            args: extra arguments
            kwargs: extra keyword arguments
        """
        super().__init__(
            parent,
            *args,
            **kwargs,
        )
        self._container_dict_class = container_dict_class
        self._container_dict = None
        self._init_ui()
        self.set_container_dict(
            container_dict=container_dict,
        )

    def set_container_dict(
        self,
        container_dict: Optional[T_ACD],
    ) -> None:
        """
        Set the container dict.

        Args:
            container_dict: a container dict
        """
        self._container_dict = container_dict
        self._remove_all_tabs()
        if container_dict is None or not isinstance(container_dict, dict):
            return
        for name, container in container_dict.items():
            entry = AttributesContainerDictEntry(
                key=name,
                value=container,
            )
            self._add_tab(
                entry=entry,
            )

    def _init_ui(self) -> None:
        layout = QHBoxLayout(self)

        self._tab_widget = self._create_tab_widget()
        self._empty_widget = self._create_empty_widget()

        self._auto_show_hide_tabs()

        layout.addWidget(self._empty_widget)
        layout.addWidget(self._tab_widget)
        layout.setMargin(0)

    def _create_empty_widget(
        self,
    ) -> QWidget:
        return self._create_buttons_widget(
            text_template=Template('Add'),
            tool_tip_template=Template('Add ${type_name}'),
            buttons_size=constants.BIG_BUTTON_SIZE,
            icons_size=constants.BIG_BUTTON_ICON_SIZE,
        )

    def _create_tab_widget(
        self,
    ) -> QTabWidget:
        corner_widget = self._create_buttons_widget(
            text_template=Template(''),
            tool_tip_template=Template('Add ${type_name}'),
            buttons_size=constants.TAB_WIDGET_CORNER_BUTTON_SIZE,
            icons_size=constants.SMALL_BUTTON_ICON_SIZE,
        )

        widget = QTabWidget(self)
        widget.setCornerWidget(corner_widget)
        widget.setTabsClosable(True)
        as_signal_instance(widget.tabCloseRequested).connect(
            self._on_tab_close_requested,
        )
        return widget

    def _create_buttons_widget(
        self,
        text_template: Template,
        tool_tip_template: Template,
        buttons_size: QSize,
        icons_size: QSize,
    ) -> QWidget:
        layout = self._create_buttons_layout(
            text_template=text_template,
            tool_tip_template=tool_tip_template,
            buttons_size=buttons_size,
            icons_size=icons_size,
        )
        margin_left, margin_top, margin_right, margin_bottom = layout.getContentsMargins()
        layout.setContentsMargins(0, margin_top, margin_right, margin_bottom)
        widget = QWidget(self)
        widget.setLayout(layout)
        return widget

    def _create_add_button(
        self,
        icon: Optional[QIcon],
        text: Optional[str],
        tool_tip: Optional[str],
        entry_args: Dict[str, Any],
        size: QSize,
        icon_size: QSize,
    ) -> QToolButton:
        widget = QToolButton(self)
        widget.setToolButtonStyle(Qt.ToolButtonTextUnderIcon)
        if text:
            widget.setText(text)
        if icon:
            widget.setIcon(icon)
        if tool_tip:
            widget.setToolTip(tool_tip)
        widget.setFixedSize(size)
        widget.setIconSize(icon_size)
        as_signal_instance(widget.clicked).connect(
            partial(
                self._on_add_button_clicked,
                entry_args,
            ),
        )
        return widget

    def _get_button_descriptions(
        self,
    ) -> List[_ButtonDescription]:
        button_descriptions: List[_ButtonDescription] = []
        item_values_type = self._container_dict_class().values_type  # type: ignore
        if isinstance(item_values_type, SubtypableMetaclass):
            values_type = cast(SubtypableMetaclass, item_values_type)
            types = values_type.get_registered_subtypes()
            for type_name, type_class in types.items():
                button_descriptions.append(
                    _ButtonDescription(
                        type_name=type_name,
                        icon_type_name=f'{values_type.__name__}/{type_class.__name__}',
                        entry_args={
                            'type': type_name,
                        },
                    ),
                )
        else:
            button_descriptions.append(
                _ButtonDescription(
                    type_name=item_values_type.__name__,
                    icon_type_name=item_values_type.__name__,
                    entry_args={},
                ),
            )
        return button_descriptions

    def _create_buttons_layout(
        self,
        text_template: Template,
        tool_tip_template: Template,
        buttons_size: QSize,
        icons_size: QSize,
    ) -> QLayout:
        layout = QHBoxLayout()
        descriptions = self._get_button_descriptions()
        for description in descriptions:
            button = self._create_add_button(
                icon=self._get_type_icon(
                    type_name=description.icon_type_name,
                ),
                text=text_template.substitute(
                    type_name=description.type_name,
                ),
                tool_tip=tool_tip_template.substitute(
                    type_name=description.type_name,
                ),
                entry_args=description.entry_args,
                size=buttons_size,
                icon_size=icons_size,
            )
            layout.addWidget(button, 1)
        layout.addStretch(1)
        return layout

    def _get_type_icon(
        self,
        type_name: str,
    ) -> Optional[QIcon]:
        paths = [
            template.substitute(
                type_name=type_name,
            )
            for template in ICON_PATH_TEMPLATES
        ]
        return _try_get_icon(
            try_paths=paths,
        )

    def _on_add_button_clicked(
        self,
        entry_args: Dict[str, Any],
    ) -> None:
        entry = self._create_container_dict_entry(
            entry_args=entry_args,
        )
        new_tab_index = self._add_tab(
            entry=entry,
        )
        self._tab_widget.setCurrentIndex(new_tab_index)

    def _remove_all_tabs(
        self,
    ) -> None:
        self._tab_widget.blockSignals(True)
        while self._tab_widget.count():
            self._tab_widget.removeTab(0)
        self._tab_widget.blockSignals(False)
        self._auto_show_hide_tabs()

    def _add_tab(
        self,
        entry: AttributesContainerDictEntry,
    ) -> int:
        from ywh2bt.gui.widgets.attribute.attributes_container_dict_entry_widget import \
            AttributesContainerDictEntryWidget  # noqa: WPS433, N400

        widget = AttributesContainerDictEntryWidget(
            entry=entry,
        )
        as_signal_instance(widget.key_changed).connect(
            self._on_container_dict_entry_key_changed,
        )
        as_signal_instance(widget.value_changed).connect(
            self._on_container_dict_entry_value_changed,
        )
        tab_index = self._tab_widget.addTab(
            widget,
            entry.key,
        )
        self._auto_show_hide_tabs()
        return tab_index

    def _on_tab_close_requested(
        self,
        tab_index: int,
    ) -> None:
        if self._container_dict:
            keys = list(self._container_dict.keys())
            self._delete_container_dict_entry(
                key=keys[tab_index],
            )
            self._tab_widget.removeTab(tab_index)
        self._auto_show_hide_tabs()

    def _auto_show_hide_tabs(
        self,
    ) -> None:
        if self._tab_widget.count() == 0:
            self._tab_widget.hide()
            self._empty_widget.show()
        else:
            self._tab_widget.show()
            self._empty_widget.hide()

    def _on_container_dict_entry_key_changed(
        self,
        entry: AttributesContainerDictEntry,
    ) -> None:
        if self._container_dict is None:
            return
        key_index = None
        for index, item in enumerate(self._container_dict.items()):
            key, value = item
            if value == entry.value:
                key_index = (key, index)
                break
        if not key_index:
            raise KeyError(f'Could not find old key for {entry}')
        self._container_dict.swap_key(
            old=key_index[0],
            new=entry.key,
        )
        self._tab_widget.setTabText(
            key_index[1],
            entry.key,
        )
        self._emit_container_dict_changed_signal()

    def _on_container_dict_entry_value_changed(
        self,
        entry: AttributesContainerDictEntry,
    ) -> None:
        if self._container_dict is None:
            return
        self._container_dict[entry.key] = entry.value
        self._emit_container_dict_changed_signal()

    def _create_container_dict_entry(
        self,
        entry_args: Dict[str, Any],
    ) -> AttributesContainerDictEntry:
        if not self._container_dict:
            self._container_dict = self._container_dict_class()  # type: ignore
        item_values_type = self._container_dict.values_type
        value = item_values_type(**entry_args)
        key = _get_next_available_key(
            container_dict=self._container_dict,
            value=value,
        )
        self._container_dict[key] = value
        self._emit_container_dict_changed_signal()
        return AttributesContainerDictEntry(
            key=key,
            value=value,
        )

    def _delete_container_dict_entry(
        self,
        key: str,
    ) -> None:
        if self._container_dict is None:
            return
        self._container_dict.pop(key)
        self._emit_container_dict_changed_signal()

    def _emit_container_dict_changed_signal(
        self,
    ) -> None:
        as_signal_instance(self.container_dict_changed).emit(
            self._container_dict,
        )


def _get_icon(
    path: str,
) -> Optional[QIcon]:
    if QFile.exists(path):
        return QIcon(path)
    return None


def _value_class_to_snake_case(
    value: Any,
) -> str:
    pattern = re.compile('(?<!^)(?=[A-Z])')
    return pattern.sub('_', value.__class__.__name__).lower()


def _try_get_icon(
    try_paths: List[str],
) -> Optional[QIcon]:
    for try_path in try_paths:
        icon = _get_icon(
            path=try_path,
        )
        if icon:
            return icon
    return None


def _get_next_available_key(
    container_dict: Optional[T_ACD],
    value: AttributesContainer,
) -> str:
    i = 1
    while True:
        snake = _value_class_to_snake_case(
            value=value,
        )
        key = f'{snake}_{i}'
        if not container_dict or key not in container_dict:
            return key
        i += 1



"""Models and functions used for attributes container dict GUI."""
from dataclasses import dataclass

from ywh2bt.core.configuration.attribute import AttributesContainer


@dataclass
class AttributesContainerDictEntry:
    """A class describing an entry of an AttributesContainerDict instance."""

    key: str
    value: AttributesContainer



"""Models and functions used for the main GUI."""
import os
import platform
import sys

import urllib3  # type: ignore
from PySide2.QtGui import QIcon
from PySide2.QtWidgets import QApplication

# noinspection PyUnresolvedReferences
import ywh2bt.gui.resources  # noqa: F401, WPS301
from ywh2bt.gui.settings import settings
from ywh2bt.gui.widgets.main_window import MainWindow
from ywh2bt.version import __VERSION__

if platform.system() == 'Windows':
    import ctypes  # noqa: WPS433

urllib3.disable_warnings(
    category=urllib3.exceptions.InsecureRequestWarning,
)


def run(*argv: str) -> None:
    """
    Start the main GUI.

    Args:
        argv: command line arguments
    """
    system = platform.system()
    if system == 'Windows':
        # set explicit AppUserModelID ; allows custom icon in Windows taskbar
        # see https://stackoverflow.com/a/1552105/248343
        ctypes.windll.shell32.SetCurrentProcessExplicitAppUserModelID(  # type: ignore
            f'yeswehack.ywh2bt.gui.{__VERSION__}'.encode('utf-8'),
        )
    elif system == 'Darwin':
        # https://stackoverflow.com/a/64856281/248343
        os.environ['QT_MAC_WANTS_LAYER'] = '1'
    args = [
        'ywh2bt-gui',
        *argv[1:],
    ]
    app = QApplication(args)

    main_window = MainWindow(
        settings=settings,
    )
    main_window.setObjectName('main_window')
    main_window.setWindowIcon(QIcon(':/resources/icons/ywh2bt.png'))
    main_window.setWindowTitle('ywh2bt-gui')
    main_window.show()
    sys.exit(app.exec_())


if __name__ == '__main__':
    run(*sys.argv)



"""Models and functions used for attributes container list GUI."""
from typing import Any, Optional, Type

from PySide2.QtCore import Signal
from PySide2.QtWidgets import QHBoxLayout, QLayout, QPushButton, QTabWidget, QVBoxLayout, QWidget

from ywh2bt.core.configuration.attribute import AttributesContainer, AttributesContainerList
from ywh2bt.gui.widgets import constants
from ywh2bt.gui.widgets.tab_widget_helper import object_to_tab_title
from ywh2bt.gui.widgets.typing import as_signal_instance

T_AC = AttributesContainer
T_ACL = AttributesContainerList[T_AC]


class AttributesContainerListWidget(QWidget):
    """Attributes container list GUI."""

    container_list_changed: Signal = Signal(AttributesContainerList)
    _container_list_class: Type[T_ACL]
    _container_list: Optional[T_ACL]

    _empty_widget: QWidget
    _tab_widget: QTabWidget

    def __init__(
        self,
        container_list_class: Type[T_ACL],
        container_list: Optional[T_ACL] = None,
        parent: Optional[QWidget] = None,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """
        Initialize self.

        Args:
            container_list_class: a concrete class extending AttributesContainerList
            container_list: an attributes container list
            parent: a parent widget
            args: extra arguments
            kwargs: extra keyword arguments
        """
        super().__init__(
            parent,
            *args,
            **kwargs,
        )
        self._container_list_class = container_list_class
        self._container_list = None
        self._init_ui()
        self.set_container_list(
            container_list=container_list,
        )

    def set_container_list(
        self,
        container_list: Optional[T_ACL],
    ) -> None:
        """
        Set the container list.

        Args:
            container_list: a container list
        """
        self._container_list = container_list
        if container_list is None:
            return
        for container in container_list:
            self._add_tab(
                container=container,
            )

    def _init_ui(self) -> None:
        layout = QHBoxLayout(self)

        self._tab_widget = self._create_tab_widget()
        self._empty_widget = self._create_empty_widget()

        layout.addWidget(self._empty_widget)
        layout.addWidget(self._tab_widget)
        layout.setMargin(0)

        self._auto_show_hide_tabs()

    def _create_empty_widget(
        self,
    ) -> QTabWidget:
        return self._create_buttons_widget()

    def _create_tab_widget(
        self,
    ) -> QTabWidget:
        widget = QTabWidget(self)
        widget.setTabsClosable(True)
        widget.setCornerWidget(self._create_buttons_widget())
        as_signal_instance(widget.tabCloseRequested).connect(
            self._on_tab_close_requested,
        )
        return widget

    def _create_buttons_widget(
        self,
    ) -> QPushButton:
        button = QPushButton(
            '+',
            self,
        )
        button.setFixedSize(constants.SMALL_BUTTON_SIZE)
        as_signal_instance(button.clicked).connect(
            self._on_add_button_clicked,
        )
        layout = QHBoxLayout()
        layout.addWidget(button, 1)
        layout.addStretch(1)
        layout.setContentsMargins(0, 0, 0, 0)
        widget = QWidget(self)
        widget.setLayout(layout)
        return widget

    def _create_buttons_layout(
        self,
        *buttons: QPushButton,
    ) -> QLayout:
        layout = QVBoxLayout()
        for button in buttons:
            layout.addWidget(button, 1)
        layout.addStretch(1)
        return layout

    def _on_add_button_clicked(
        self,
        checked: bool,
    ) -> None:
        new_entry = self._create_container_list_entry()
        new_tab_index = self._add_tab(
            container=new_entry,
        )
        self._tab_widget.setCurrentIndex(new_tab_index)

    def _auto_show_hide_tabs(
        self,
    ) -> None:
        if self._tab_widget.count() == 0:
            self._tab_widget.hide()
            self._empty_widget.show()
        else:
            self._tab_widget.show()
            self._empty_widget.hide()

    def _add_tab(
        self,
        container: AttributesContainer,
    ) -> int:
        from ywh2bt.gui.widgets.attribute.attributes_container_widget import AttributesContainerWidget  # noqa: WPS433
        widget = AttributesContainerWidget(
            parent=self,
            container_class=container.__class__,
            container=container,
        )
        as_signal_instance(widget.container_changed).connect(
            self._on_attributes_container_widget_changed,
        )
        title = self._attributes_container_to_tab_title(
            container=container,
        )
        tab_index = self._tab_widget.addTab(
            widget,
            title,
        )
        self._auto_show_hide_tabs()
        return tab_index

    def _on_tab_close_requested(
        self,
        tab_index: int,
    ) -> None:
        self._delete_container_list_entry(
            index=tab_index,
        )
        self._tab_widget.removeTab(tab_index)

        self._auto_show_hide_tabs()

    def _attributes_container_to_tab_title(
        self,
        container: AttributesContainer,
    ) -> str:
        if self._container_list is None:
            return 'Tab'
        title = object_to_tab_title(
            obj=container,
        )
        if not title:
            index = self._container_list.index(container)
            class_name = self._container_list.values_type.__name__
            position = index + 1
            title = f'{class_name} #{position}'
        return title

    def _on_attributes_container_widget_changed(
        self,
        container: AttributesContainer,
    ) -> None:
        if self._container_list is None:
            return
        index = self._container_list.index(container)
        self._tab_widget.setTabText(
            index,
            self._attributes_container_to_tab_title(
                container=container,
            ),
        )
        self._emit_container_list_changed_signal()

    def _create_container_list_entry(
        self,
    ) -> AttributesContainer:
        if self._container_list is None:
            self._container_list = self._container_list_class()  # type: ignore
        item_values_type = self._container_list.values_type
        entry = item_values_type()
        self._container_list.append(entry)
        self._emit_container_list_changed_signal()
        return entry

    def _delete_container_list_entry(
        self,
        index: int,
    ) -> None:
        if self._container_list is None:
            return
        self._container_list.pop(index)
        self._emit_container_list_changed_signal()

    def _emit_container_list_changed_signal(
        self,
    ) -> None:
        as_signal_instance(self.container_list_changed).emit(
            self._container_list,
        )



"""Models and functions used for root configuration GUI."""
from datetime import datetime
from io import StringIO
from typing import Any, Optional, Tuple, cast

from PySide2.QtCore import QFileInfo, QSettings, Signal
from PySide2.QtGui import QFontDatabase, QFontMetrics, Qt
from PySide2.QtWidgets import (
    QFileDialog,
    QMessageBox,
    QPlainTextEdit,
    QProgressBar,
    QScrollArea,
    QSplitter,
    QTabWidget,
    QVBoxLayout,
    QWidget,
)

from ywh2bt.core.configuration.root import RootConfiguration
from ywh2bt.core.core import AVAILABLE_FORMATS, get_root_configuration_loader
from ywh2bt.core.error import print_error
from ywh2bt.core.exceptions import CoreException
from ywh2bt.core.loader import RootConfigurationLoader
from ywh2bt.gui.dialog.error import ErrorDialogMixin
from ywh2bt.gui.dialog.file import FileFormatDialogFilters
from ywh2bt.gui.hashing import file_checksum
from ywh2bt.gui.widgets.attribute.attributes_container_widget import AttributesContainerWidget
from ywh2bt.gui.widgets.logs_widget import LogEntry, LogType, LogsWidget
from ywh2bt.gui.widgets.root_configuration_entry import RootConfigurationEntry, RootConfigurationEntryFile
from ywh2bt.gui.widgets.thread.synchronizer import SynchronizerThread
from ywh2bt.gui.widgets.thread.tester import TesterThread
from ywh2bt.gui.widgets.typing import as_signal_instance


class RootConfigurationWidget(QWidget, ErrorDialogMixin):
    """Root configuration GUI."""

    root_configuration_modified: Signal = Signal(RootConfigurationEntry)
    root_configuration_saved: Signal = Signal(RootConfigurationEntry)
    root_configuration_saved_as: Signal = Signal(str, str)

    _log_entry_available: Signal = Signal(LogEntry)

    _file_format_dialog_filters: FileFormatDialogFilters
    _settings: QSettings

    _entry: RootConfigurationEntry

    _tab_widget: QTabWidget
    _edit_tab_widget: AttributesContainerWidget
    _raw_widget: QPlainTextEdit

    _logs_widget: LogsWidget
    _progress_bar: QProgressBar
    _thread_running: bool

    def __init__(
        self,
        settings: QSettings,
        entry: RootConfigurationEntry,
        parent: Optional[QWidget] = None,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """
        Initialize self.

        Args:
            settings: application settings
            entry: a root configuration entry
            parent: a parent widget
            args: extra arguments
            kwargs: extra keyword arguments
        """
        super().__init__(
            parent,
            *args,
            **kwargs,
        )
        self._entry = entry
        self._settings = settings
        self._file_format_dialog_filters = FileFormatDialogFilters(
            formats=AVAILABLE_FORMATS,
        )
        self._init_ui()
        self._on_entry_changed()
        as_signal_instance(self._log_entry_available).connect(
            self._on_log_entry_available,
        )
        self._thread_running = False

    def set_entry(
        self,
        entry: RootConfigurationEntry,
    ) -> None:
        """
        Set the entry.

        Args:
            entry: an entry
        """
        self._entry = entry
        self._on_entry_changed()

    def _on_entry_changed(
        self,
        update_raw: bool = True,
    ) -> None:
        self._edit_tab_widget.set_container(
            container=self._entry.configuration,
        )
        self._edit_tab_widget.setEnabled(self._entry.configuration is not None)
        self._validate_entry()
        if update_raw:
            self._update_raw_widget()

    def _init_ui(
        self,
    ) -> None:
        layout = QVBoxLayout(self)

        self._tab_widget = QTabWidget(self)

        self._edit_tab_widget = self._create_edit_tab_widget()
        self._tab_widget.addTab(
            self._wrap_in_scroll_area(
                widget=self._edit_tab_widget,
            ),
            'Configuration',
        )
        raw_tab_widget = self._create_raw_tab_widget()
        self._tab_widget.addTab(
            raw_tab_widget,
            f'Raw ({self._entry.raw_format})',
        )

        self._logs_widget = LogsWidget(self)

        self._progress_bar = self._create_progress_bar()

        splitter = self._create_splitter()

        layout.addWidget(splitter)

    def _create_edit_tab_widget(
        self,
    ) -> AttributesContainerWidget:
        widget = AttributesContainerWidget(
            parent=self,
            container_class=RootConfiguration,
            container=self._entry.configuration,
        )
        as_signal_instance(widget.container_changed).connect(
            self._on_root_configuration_modified,
        )
        return widget

    def _create_progress_bar(
        self,
    ) -> QProgressBar:
        widget = QProgressBar(self)
        widget.setRange(0, 100)
        widget.setFixedHeight(15)
        return widget

    def _create_splitter(
        self,
    ) -> QSplitter:
        splitter = QSplitter(
            Qt.Vertical,
            self,
        )
        splitter.addWidget(self._tab_widget)
        splitter.addWidget(self._logs_widget)
        splitter.addWidget(self._progress_bar)
        splitter.setCollapsible(0, False)  # noqa: WPS425
        splitter.setCollapsible(1, False)  # noqa: WPS425
        splitter.setCollapsible(2, False)  # noqa: WPS425
        splitter.setStretchFactor(0, 98)
        splitter.setStretchFactor(1, 1)
        splitter.setStretchFactor(2, 1)
        return splitter

    def _wrap_in_scroll_area(
        self,
        widget: QWidget,
    ) -> QScrollArea:
        scroll_area = QScrollArea(self)
        scroll_area.setWidget(widget)
        scroll_area.setWidgetResizable(True)
        return scroll_area

    def _on_root_configuration_modified(
        self,
        configuration: RootConfiguration,
    ) -> None:
        self._entry.configuration = configuration
        self._validate_entry()
        self._update_raw_widget()
        as_signal_instance(self.root_configuration_modified).emit(
            self._entry,
        )

    def _create_raw_tab_widget(
        self,
    ) -> QWidget:
        layout = QVBoxLayout()
        self._raw_widget = self._create_raw_widget()
        layout.addWidget(self._raw_widget, 1)

        widget = QWidget(self)
        widget.setLayout(layout)
        return widget

    def _create_raw_widget(
        self,
    ) -> QWidget:
        widget = QPlainTextEdit(self)
        font = QFontDatabase.systemFont(QFontDatabase.FixedFont)
        font.setFixedPitch(True)
        metrics = QFontMetrics(font)
        widget.setFont(font)
        widget.setTabStopDistance(2 * metrics.width(' '))
        as_signal_instance(widget.textChanged).connect(
            self._on_raw_changed,
        )
        return widget

    def _on_raw_changed(
        self,
    ) -> None:
        raw = self._raw_widget.toPlainText()
        try:
            self._entry.raw = raw
        except CoreException as e:
            self._log(
                log_type=LogType.error,
                message=_format_error(
                    error=e,
                ),
            )
        self._on_entry_changed(
            update_raw=False,
        )
        as_signal_instance(self.root_configuration_modified).emit(
            self._entry,
        )

    def _validate_entry(
        self,
    ) -> None:
        if not self._entry.configuration:
            return
        try:
            self._entry.configuration.validate()
        except CoreException as e:
            self._log(
                log_type=LogType.error,
                message=_format_error(
                    error=e,
                ),
            )
        else:
            self._log(
                log_type=LogType.success,
                message='Valid',
            )

    def _on_log_entry_available(
        self,
        log_entry: LogEntry,
    ) -> None:
        self._logs_widget.add_log_entry(
            entry=log_entry,
        )

    def _log(
        self,
        message: str,
        log_type: LogType = LogType.standard,
    ) -> None:
        self._logs_widget.add_log_entry(
            entry=LogEntry(
                date_time=datetime.now(),
                context=self._entry.name,
                message=message,
                log_type=log_type,
            ),
        )

    def _get_loader(
        self,
    ) -> RootConfigurationLoader:
        return get_root_configuration_loader(
            file_format=self._entry.raw_format,
        )

    def _update_raw_widget(
        self,
    ) -> None:
        plain_text = self._entry.raw
        self._raw_widget.blockSignals(True)
        cursor = self._raw_widget.textCursor()
        position = cursor.position()
        self._raw_widget.setPlainText(plain_text)
        cursor.setPosition(min(position, len(plain_text)))
        self._raw_widget.setTextCursor(cursor)
        self._raw_widget.blockSignals(False)

    def _get_last_opened_dir(
        self,
    ) -> str:
        return QFileInfo(
            cast(str, self._settings.value('last_opened_file')),
        ).dir().path()

    def save(
        self,
    ) -> None:
        """Engage the saving process."""
        entry = self._entry
        save_invalid = self._prompt_save_invalid_entry(
            entry=entry,
        )
        if not save_invalid:
            return
        path_format = self._prompt_save_path_format(
            entry=entry,
        )
        if not path_format:
            return
        file_path, file_format = path_format
        if entry.configuration is None:
            entry.configuration = RootConfiguration()
        saved = self._save_to_file(
            entry=entry,
            file_path=file_path,
            file_format=file_format,
        )
        if not saved:
            return
        self._setup_entry_from_saved_file(
            file_info=QFileInfo(file_path),
            file_format=file_format,
        )
        self._log(
            log_type=LogType.success,
            message=f'{self._entry.name} saved',
        )

    def save_as(
        self,
    ) -> None:
        """Engage the saving process into a new file."""
        entry = self._entry
        save_invalid = self._prompt_save_invalid_entry(
            entry=entry,
        )
        if not save_invalid:
            return
        path_format = self._prompt_new_file(
            format_filters=self._file_format_dialog_filters.get_filters_string(),
        )
        if not path_format:
            return
        file_path, file_format = path_format
        if entry.configuration is None:
            entry.configuration = RootConfiguration()
        saved = self._save_to_file(
            entry=entry,
            file_path=file_path,
            file_format=file_format,
        )
        if not saved:
            return
        as_signal_instance(self.root_configuration_saved_as).emit(
            file_path,
            file_format,
        )

    def _setup_entry_from_saved_file(
        self,
        file_info: QFileInfo,
        file_format: str,
    ) -> None:
        self._entry.file = RootConfigurationEntryFile(
            info=file_info,
            file_format=file_format,
            checksum=file_checksum(file_info.filePath()),
        )
        self._entry.name = file_info.fileName()
        self._entry.original_raw = self._entry.raw
        as_signal_instance(self.root_configuration_modified).emit(
            self._entry,
        )
        as_signal_instance(self.root_configuration_saved).emit(
            self._entry,
        )

    def test(
        self,
    ) -> None:
        """Test the configuration."""
        entry = self._entry
        if self._thread_running or not entry.configuration or not entry.configuration.is_valid():
            return
        thread = TesterThread(
            parent=self,
            entry=entry,
            log_entry_available_signal=as_signal_instance(self._log_entry_available),
        )
        as_signal_instance(thread.test_started).connect(
            self._on_thread_started,
        )
        as_signal_instance(thread.test_ended).connect(
            self._on_thread_ended,
        )
        thread.start()

    def synchronize(
        self,
    ) -> None:
        """Synchronize the configuration."""
        entry = self._entry
        if self._thread_running or not entry.configuration or not entry.configuration.is_valid():
            return
        thread = SynchronizerThread(
            parent=self,
            entry=entry,
            log_entry_available_signal=as_signal_instance(self._log_entry_available),
        )
        as_signal_instance(thread.synchronization_started).connect(
            self._on_thread_started,
        )
        as_signal_instance(thread.synchronization_ended).connect(
            self._on_thread_ended,
        )
        thread.start()

    def _on_thread_started(
        self,
    ) -> None:
        self._thread_running = True
        self._progress_bar.setRange(0, 0)

    def _on_thread_ended(
        self,
    ) -> None:
        self._progress_bar.setRange(0, 100)
        self._progress_bar.reset()
        self._thread_running = False

    def _prompt_save_invalid_entry(
        self,
        entry: RootConfigurationEntry,
    ) -> bool:
        error_message = None
        if entry.configuration is None:
            error_message = 'Are you sure you want to save an empty configuration?'
        elif not entry.configuration.is_valid():
            error_message = 'Are you sure you want to save an invalid configuration?'
        if error_message:
            reply = QMessageBox.question(
                self,
                f'Save {entry.name}',
                error_message,
                QMessageBox.Yes | QMessageBox.No,
                QMessageBox.No,
            )
            if reply == QMessageBox.No:
                return False
        return True

    def _prompt_save_path_format(
        self,
        entry: RootConfigurationEntry,
    ) -> Optional[Tuple[str, str]]:
        if entry.file:
            exists = entry.file.info.exists()
            changed = entry.file.checksum != file_checksum(entry.file.info.filePath())
            if exists and changed:
                overwrite = self._prompt_overwrite_changed_file(
                    entry=entry,
                )
                if not overwrite:
                    return None
            file_path = entry.file.info.filePath()
            file_format = entry.file.file_format
        else:
            new_file = self._prompt_new_file(
                format_filters=self._file_format_dialog_filters.get_filters_string_for(
                    format_name=self._entry.raw_format,
                ) or '',
            )
            if not new_file:
                return None
            file_path, file_format = new_file
        return file_path, file_format

    def _prompt_overwrite_changed_file(
        self,
        entry: RootConfigurationEntry,
    ) -> bool:
        if not entry.file:
            return False
        reply = QMessageBox.question(
            self,
            f'Overwrite {entry.name}',
            f'{entry.file.info.filePath()} has changed on the disk.\nOverwrite?',
            QMessageBox.Yes | QMessageBox.No,
            QMessageBox.No,
        )
        return reply == QMessageBox.Yes

    def _prompt_new_file(
        self,
        format_filters: str,
    ) -> Optional[Tuple[str, str]]:
        last_opened_dir = self._get_last_opened_dir()
        file_path, chosen_filter = QFileDialog.getSaveFileName(
            self,
            'New file',
            last_opened_dir,
            format_filters or '',
        )
        if not file_path or not chosen_filter:
            return None
        file_format = self._file_format_dialog_filters.get_format_name(
            filter_string=chosen_filter,
        )
        if not file_format:
            return None
        return file_path, file_format

    def _save_to_file(
        self,
        entry: RootConfigurationEntry,
        file_path: str,
        file_format: str,
    ) -> bool:
        if not entry.configuration:
            return False
        loader = self._get_root_configuration_loader(
            file_format=file_format,
        )
        if not loader:
            return False
        try:
            with open(file_path, 'w') as destination_file:
                loader.save(
                    data=entry.configuration,
                    stream=destination_file,
                )
        except (IOError, CoreException) as save_error:
            self.show_exception_dialog(
                parent=self,
                text='Save error',
                informative_text=f'Unable to save file:\n{file_path}',
                exception=save_error,
            )
            return False
        return True

    def _get_root_configuration_loader(
        self,
        file_format: str,
    ) -> Optional[RootConfigurationLoader]:
        try:
            return get_root_configuration_loader(
                file_format=file_format,
            )
        except CoreException as loader_error:
            self.show_exception_dialog(
                parent=self,
                text='Loader error',
                informative_text=f'Unable to get loader for {file_format}',
                exception=loader_error,
            )
        return None


def _format_error(
    error: Exception,
) -> str:
    error_stream = StringIO()
    print_error(
        stream=error_stream,
        error=error,
    )
    return error_stream.getvalue()



"""Models and functions used for formatting reports data."""
from string import Template
from typing import Dict, Tuple, cast

from bs4 import BeautifulSoup, Tag  # type: ignore

from ywh2bt.core.api.formatter.formatter import ReportMessageFormatter
from ywh2bt.core.converter.html2jira import html2jira

REPORT_TITLE_TEMPLATE = '${local_id} : ${title}'
REPORT_DESCRIPTION_TEMPLATE = """
|| Title || ${local_id} : ${title} ||
| Priority | ${priority_name} |
| ${bug_type_label} | [${bug_type_name}|${bug_type_link}] => [Remediation|${bug_type_remediation_link}] |
| ${scope_label} | {noformat}${scope}{noformat} |
| Severity | ${cvss_criticity}, score: ${cvss_score}, vector: {noformat}${cvss_vector}{noformat} |
| ${end_point_label} | ${end_point} |
| ${vulnerable_part_label} | ${vulnerable_part} |
| ${part_name_label} | ${part_name} |
| ${payload_sample_label} | ${payload_sample} |
| Technical Environment | ${technical_environment} |

${description}
"""
CLOSE_LOG_TEMPLATE = """
*Report closed*:

${old_status} -> ${new_status}

*Comment*:

${comment}
"""
COMMENT_BODY_TEMPLATE = """
*Date*: ${date}

${body}
"""
COMMENT_LOG_TEMPLATE = """
*Comment*:

${comment}
"""
CVSS_UPDATE_LOG_TEMPLATE = """
*CVSS updated*:

|| Detail || Old value || New value ||
| *Severity* | ${old_cvss_criticity} | ${new_cvss_criticity} |
| *Score* | ${old_cvss_score} | ${new_cvss_score} |
"""
STATUS_UPDATE_LOG_TEMPLATE = """
*Status updated*:

${old_status} -> ${new_status}

*Comment*:

${comment}
"""
FIX_VERIFIED_LOG_TEMPLATE = """
*Fix verification*:

${verified}

*Comment*:

${comment}
"""
DETAILS_UPDATE_LOG_TEMPLATE = """
*Details updated*:

|| Detail || Old value || New value ||
${details_lines}
"""
DETAILS_UPDATE_LOG_LINE_TEMPLATE = """| *${updated_property}* | {noformat}${old_value}{noformat} | {noformat}${new_value}{noformat} |
"""
PRIORITY_UPDATE_LOG_TEMPLATE = """
*Priority updated*:

${new_priority}
"""
REWARD_LOG_TEMPLATE = """
*Comment from reward*:

${comment}
"""


class JiraReportMessageFormatter(ReportMessageFormatter):
    """A report formatter to JIRA."""

    _code_block_template = Template('{code${language}}\n${content}\n{code}')

    _code_block_languages = [
        'actionscript',
        'ada',
        'applescript',
        'bash',
        'c',
        'c#',
        'c++',
        'cpp',
        'css',
        'erlang',
        'go',
        'groovy',
        'haskell',
        'html',
        'java',
        'javascript',
        'js',
        'json',
        'lua',
        'none',
        'nyan',
        'objc',
        'perl',
        'php',
        'python',
        'r',
        'rainbow',
        'ruby',
        'scala',
        'sh',
        'sql',
        'swift',
        'visualbasic',
        'xml',
        'yaml',
    ]

    def __init__(self) -> None:
        """Initialize self."""
        super().__init__(
            report_title_template=Template(REPORT_TITLE_TEMPLATE),
            report_description_template=Template(REPORT_DESCRIPTION_TEMPLATE),
            close_log_template=Template(CLOSE_LOG_TEMPLATE),
            comment_body_template=Template(COMMENT_BODY_TEMPLATE),
            comment_log_template=Template(COMMENT_LOG_TEMPLATE),
            cvss_update_log_template=Template(CVSS_UPDATE_LOG_TEMPLATE),
            status_update_log_template=Template(STATUS_UPDATE_LOG_TEMPLATE),
            fix_verified_log_template=Template(FIX_VERIFIED_LOG_TEMPLATE),
            details_update_log_template=Template(DETAILS_UPDATE_LOG_TEMPLATE),
            details_update_log_line_template=Template(DETAILS_UPDATE_LOG_LINE_TEMPLATE),
            priority_update_log_template=Template(PRIORITY_UPDATE_LOG_TEMPLATE),
            reward_log_template=Template(REWARD_LOG_TEMPLATE),
        )

    def transform_report_description_html(
        self,
        description_html: str,
    ) -> str:
        """
        Transform the report description.

        Args:
            description_html: a report description in HTML

        Returns:
            a transformed description
        """
        return self._html2jira(
            html=description_html,
        )

    def transform_html(
        self,
        html: str,
    ) -> str:
        """
        Transform from HTML.

        Args:
            html: HTML

        Returns:
            HTML transformed to JIRA
        """
        return self._html2jira(
            html=html,
        )

    def _html2jira(
        self,
        html: str,
    ) -> str:
        body_html = self._img_tags_to_jira_tags(
            html=html,
        )
        body_html, code_blocks = self._extract_code_blocks(
            html=body_html,
        )
        body_jira = html2jira(
            html=body_html,
        )
        for block_id, code_block in code_blocks.items():
            body_jira = body_jira.replace(
                block_id,
                code_block,
            )
        return body_jira

    def _img_tags_to_jira_tags(
        self,
        html: str,
    ) -> str:
        soup = BeautifulSoup(
            html,
            features='lxml',
        )
        clean_html = str(soup)
        for img in soup.findAll('img'):
            alt = img.attrs.get('alt', '')
            src = img.attrs.get('src', '')
            a = f'!{alt}|{src}!'
            clean_html = clean_html.replace(str(img), a)
        return clean_html

    def _extract_code_blocks(
        self,
        html: str,
    ) -> Tuple[str, Dict[str, str]]:
        return self._extract_code_blocks_from_soup(
            soup=BeautifulSoup(
                html,
                features='lxml',
            ),
        )

    def _extract_code_blocks_from_soup(
        self,
        soup: BeautifulSoup,
    ) -> Tuple[str, Dict[str, str]]:
        clean_html = str(soup)
        code_blocks = {}
        for idx, code_tag in enumerate(soup.findAll('code')):
            block_id = f'{{yeswehack_code_section_{idx}}}'
            language = self._extract_code_block_language(
                code_tag=code_tag,
            )
            code_blocks[block_id] = self._code_block_template.substitute(
                language=f':{language}' if language else '',
                content=''.join(map(str, code_tag.contents)),
            )
            clean_html = clean_html.replace(str(code_tag), block_id)
        return clean_html, code_blocks

    def _extract_code_block_language(
        self,
        code_tag: Tag,
    ) -> str:
        title = code_tag.attrs.get('class', '')
        if isinstance(title, list):
            title = title[0] if title else ''
        language = cast(str, title.lower().replace('language-', ''))
        return language if language in self._code_block_languages else ''



"""Models and functions used for JSON schema."""
import json
from io import StringIO
from types import MappingProxyType
from typing import Any, Dict, Mapping, Optional, Type

from typing_extensions import Protocol

from ywh2bt.core.configuration.attribute import (
    Attribute,
    AttributesContainer,
    AttributesContainerDict,
    AttributesContainerList,
    ExportableDict,
    ExportableList,
)
from ywh2bt.core.configuration.root import RootConfiguration
from ywh2bt.core.configuration.subtypable import SubtypableMetaclass
from ywh2bt.core.schema.error import SchemaError

Json = Dict[str, Any]


def root_configuration_as_json_schema() -> str:
    """
    Get a JSON schema for RootConfiguration.

    Raises:
        SchemaError: if an error occurred during the creation of the schema

    Returns:
        the JSON schema
    """
    schema = _root_configuration_to_schema()
    stream = StringIO()
    try:
        json.dump(
            schema,
            fp=stream,
            indent=2,
        )
    except TypeError as dump_error:
        raise SchemaError('JSON dump error') from dump_error
    return stream.getvalue()


def _root_configuration_to_schema() -> Json:
    schema = {
        '$schema': 'http://json-schema.org/draft-07/schema',
        'title': 'Root configuration',
        'description': 'Root configuration for ywh2bt synchronization',
        'definitions': {},
    }
    schema.update(
        _attributes_container_to_schema(
            root_schema=schema,
            attribute=Attribute.create(
                value_type=RootConfiguration,
            ),
            as_ref=False,
        ),
    )
    return schema


def _attribute_to_schema(
    root_schema: Json,
    attribute: Attribute[Any],
    extra_properties: Optional[Json] = None,
) -> Json:
    value_type = attribute.value_type
    for known_type, protocol in type_to_schema_protocols.items():
        if issubclass(value_type, known_type):
            return protocol(
                root_schema=root_schema,
                attribute=attribute,
                extra_properties=extra_properties,
                as_ref=True,
            )
    raise SchemaError(f'Unhandled {repr(value_type.__name__)}')


def _attribute_to_base_schema(
    attribute: Attribute[Any],
) -> Json:
    schema = {}
    if attribute.short_description:
        schema['title'] = attribute.short_description
    if attribute.description:
        schema['description'] = attribute.description
    if attribute.deprecated:
        description = schema.get('description')
        if description:
            schema['description'] = f'(Deprecated)\n{description}'
        else:
            schema['description'] = '(Deprecated)'
    if attribute.default is not None:
        schema['default'] = attribute.default
    return schema


def _attributes_container_to_schema(
    root_schema: Json,
    attribute: Attribute[AttributesContainer],
    extra_properties: Optional[Json] = None,
    as_ref: bool = False,
) -> Json:
    required = []
    value_type = attribute.value_type
    properties = {}
    if extra_properties:
        properties.update(extra_properties)
    for name, child_attribute in value_type.get_attributes().items():
        if child_attribute.required:
            required.append(name)
        properties[child_attribute.name] = {
            **_attribute_to_base_schema(
                attribute=child_attribute,
            ),
            **_attribute_to_schema(
                root_schema=root_schema,
                attribute=child_attribute,
            ),
        }

    schema: Json = {
        'type': 'object',
    }
    if properties:
        schema['properties'] = properties
    if required:
        schema['required'] = required
    schema['additionalProperties'] = False

    if not as_ref:
        return schema

    root_schema['definitions'][value_type.__name__] = schema
    return {
        '$ref': f'#/definitions/{value_type.__name__}',
    }


def _attribute_container_dict_to_schema(
    root_schema: Json,
    attribute: Attribute[AttributesContainerDict[Any]],
    extra_properties: Optional[Json] = None,
    as_ref: bool = False,
) -> Json:
    return {
        'type': 'object',
        'patternProperties': {
            '.+': _get_pattern_properties(
                root_schema=root_schema,
                attribute=attribute,
            ),
        },
        'additionalProperties': False,
    }


def _get_pattern_properties(
    root_schema: Json,
    attribute: Attribute[Any],
) -> Json:
    value_type = attribute.value_type
    instance = value_type()
    item_values_type = instance.values_type
    if isinstance(item_values_type, SubtypableMetaclass):
        any_of = []
        for type_name, type_class in item_values_type.get_registered_subtypes().items():
            any_of.append(
                _attribute_to_schema(
                    root_schema=root_schema,
                    attribute=Attribute.create(
                        value_type=type_class,
                    ),
                    extra_properties={
                        'type': {
                            'const': type_name,
                        },
                    },
                ),
            )
        return {
            'anyOf': any_of,
        }
    return _attribute_to_schema(
        root_schema=root_schema,
        attribute=Attribute.create(
            value_type=item_values_type,
        ),
    )


def _attribute_container_list_to_schema(
    root_schema: Json,
    attribute: Attribute[AttributesContainerList[Any]],
    extra_properties: Optional[Json] = None,
    as_ref: bool = False,
) -> Json:
    value_type = attribute.value_type
    instance = value_type()  # type: ignore
    item_values_type = instance.values_type
    return {
        'type': 'array',
        'items': _attribute_to_schema(
            root_schema=root_schema,
            attribute=Attribute.create(
                value_type=item_values_type,
            ),
        ),
    }


def _exportable_dict_to_schema(
    root_schema: Json,
    attribute: Attribute[ExportableDict[Any, Any, Any, Any]],
    extra_properties: Optional[Json] = None,
    as_ref: bool = False,
) -> Json:
    return {
        'type': 'object',
        'patternProperties': {
            '.*': {
                'type': 'string',
            },
        },
    }


def _exportable_list_to_schema(
    root_schema: Json,
    attribute: Attribute[ExportableList[Any, Any]],
    extra_properties: Optional[Json] = None,
    as_ref: bool = False,
) -> Json:
    return {
        'type': 'array',
        'items': {
            'type': 'string',
        },
    }


def _str_to_schema(
    root_schema: Json,
    attribute: Attribute[str],
    extra_properties: Optional[Json] = None,
    as_ref: bool = False,
) -> Json:
    return {
        'type': 'string',
    }


def _bool_to_schema(
    root_schema: Json,
    attribute: Attribute[bool],
    extra_properties: Optional[Json] = None,
    as_ref: bool = False,
) -> Json:
    return {
        'type': 'boolean',
    }


class _ToSchemaProtocol(Protocol):

    def __call__(
        self,
        root_schema: Json,
        attribute: Attribute[Any],
        extra_properties: Optional[Json] = None,
        as_ref: bool = False,
    ) -> Json:
        ...  # noqa: WPS428


TrackerClientClassesType = Mapping[
    Type[Any],
    _ToSchemaProtocol,
]
type_to_schema_protocols: TrackerClientClassesType = MappingProxyType({
    AttributesContainer: _attributes_container_to_schema,
    AttributesContainerDict: _attribute_container_dict_to_schema,
    AttributesContainerList: _attribute_container_list_to_schema,
    ExportableDict: _exportable_dict_to_schema,
    ExportableList: _exportable_list_to_schema,
    str: _str_to_schema,
    bool: _bool_to_schema,
})



"""Models and functions used for states encryption."""
import base64
import json
from typing import Any, Dict, Tuple

from ywh2bt.core.crypt.encrypt import Encryptor
from ywh2bt.core.crypt.error import CryptError
from ywh2bt.core.crypt.key import Key
from ywh2bt.core.state.error import StateError
from ywh2bt.core.state.state import State


class StateEncryptor:
    """A simple state encryptor."""

    @classmethod
    def encrypt(
        cls,
        state: State,
        key: str,
    ) -> str:
        """
        Encrypt a state with a key.

        Args:
            state: a state
            key: a cryptographic key

        Raises:
            StateError: if the state could not be encrypted

        Returns:
            The encrypted state.
        """
        encryptor = Encryptor(
            key=Key.from_str(
                key=key,
            ),
        )
        try:
            encrypted_data = encryptor.encrypt(
                message_bytes=cls._as_bytes(
                    state=state,
                ),
            )
        except CryptError as e:
            raise StateError(f'Unable to encrypt state {state}') from e
        encrypted_base64 = base64.b64encode(
            encrypted_data,
        )
        encrypted = encrypted_base64.decode('utf-8')
        return f'[YWH2BT:S:{encrypted}]'

    @classmethod
    def _as_bytes(
        cls,
        state: State,
    ) -> bytes:
        json_string = cls._as_json_string(
            state=state,
        )
        return json_string.encode('utf-8')

    @classmethod
    def _as_json_string(
        cls,
        state: State,
    ) -> str:
        try:
            return json.dumps(
                cls._as_meta(
                    state=state,
                ),
            )
        except (TypeError, ValueError) as e:
            raise StateError(f'Unable to serialize state {state}') from e

    @classmethod
    def _as_meta(
        cls,
        state: State,
    ) -> Tuple[str, Dict[str, Any]]:
        return (
            state.__class__.__name__,
            state.as_dict(),
        )
